"""
DEPRECIATED: Use `subgrounds.contrib.plotly` instead
"""

import warnings

from subgrounds.contrib.plotly import (
    Bar,
    Barpolar,
    Box,
    Candlestick,
    Carpet,
    Choropleth,
    Choroplethmapbox,
    Contour,
    Contourcarpet,
    Densitymapbox,
    Figure,
    Funnel,
    Heatmap,
    Histogram,
    Histogram2d,
    Histogram2dContour,
    Icicle,
    Indicator,
    Ohlc,
    Parcats,
    Parcoords,
    Pie,
    Sankey,
    Scatter,
    Scatter3d,
    Scattercarpet,
    Scattergeo,
    Scattermapbox,
    Scatterpolar,
    Sunburst,
    Surface,
    Table,
    TraceWrapper,
    Treemap,
    Violin,
    Waterfall,
)

__all__ = [
    "Figure",
    "TraceWrapper",
    "Scatter",
    "Pie",
    "Bar",
    "Heatmap",
    "Contour",
    "Table",
    "Box",
    "Violin",
    "Histogram",
    "Histogram2d",
    "Histogram2dContour",
    "Ohlc",
    "Candlestick",
    "Waterfall",
    "Funnel",
    "Indicator",
    "Scatter3d",
    "Surface",
    "Scattergeo",
    "Choropleth",
    "Scattermapbox",
    "Choroplethmapbox",
    "Densitymapbox",
    "Scatterpolar",
    "Barpolar",
    "Sunburst",
    "Treemap",
    "Icicle",
    "Sankey",
    "Parcoords",
    "Parcats",
    "Carpet",
    "Scattercarpet",
    "Contourcarpet",
]


warnings.warn(
    "Importing from `subgrounds.plotly_wrappers` is deprecated."
    " Use `subgrounds.contrib.plotly` instead.\n"
    "Will be removed in a future version.",
    DeprecationWarning,
)



""" Subgrounds pagination strategies

This module implements functions and data structures used to implement the
two ``PaginationStrategies`` that Subgrounds offers.

For both strategies, pagination is done in four steps:

#. Generate one or many :class:`PaginationNode` objects per query document.
   These are tree-like data structures that extract all information about pagination
   fields (i.e.: fields that return lists of entities) while maintaining the nestedness
   relationship between each pagination field (i.e.: which is nested in which).

#. The query document is normalized such that every pagination field in the query has:
   
    #. An ordering (i.e.: ``orderBy`` and ``orderDirection`` are specified)
    #. A ``first`` argument set to the ``firstN`` variable
    #. A ``skip`` argument set to the ``skipN`` variable
    #. A ``where`` filter with the filter name derived from the ordering and the
       value being a variable named ``lastOrderingValueN``
 
   In other words, the query will be transformed in a form which allows Subgrounds
   to paginate automatically by simply setting the set of pagination variables
   (i.e.: ``firstN``, ``skipN`` and ``lastOrderingValueN``) to different
   values. Each field that requires pagination (i.e.: each field that yields a list)
   will have its own set of variables, hence the ``N`` post-fix.
 
   Example:
   The initial query
 
   .. code-block:: none
 
     query {
       items(
         orderBy: timestamp,
         orderDirection: desc,
         first: 10000
       ) {
         foo
       }
     }
 
   will be transformed to
 
   .. code-block:: none
 
     query($first0: Int, $skip0: Int, $lastOrderingValue0: BigInt) {
       items(
         orderBy: timestamp,
         orderDirection: desc,
         first: $first0,
         skip: $skip0,
         where: {
           timestamp_lt: $lastOrderingValue0
         }
       ) {
         foo
       }
     }

#. For each data page, generate the values for the pagination variables (i.e.:
   ``firstN``, ``skipN`` and ```lastOrderingValueN```)

#. If some variables are undefined (i.e.: they are present in the query document,
   but not given a value as part of step 3), then the document is pruned and all
   selections (and sub-selections) containing undefined variables are removed.

Depending on the strategy, the variable values computed at step 3 will change.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from functools import partial
from itertools import count
from typing import Any, Callable, Iterator, Literal, Optional, Protocol

from pipe import map, traverse

from subgrounds.query import Document
from subgrounds.schema import SchemaMeta
from subgrounds.utils import extract_data

from .preprocess import (
    PaginationNode,
    generate_pagination_nodes,
    normalize,
    prune_doc,
)

PAGE_SIZE = 900


class StopPagination(Exception):
    """Raise within :class:`PaginationStategy` when pagination is completed, usually
    within :func:`~subgrounds.pagination.pagination.PaginationStrategy.step`.

    Causes :func:`~subgrounds.pagination.pagination.paginate` to be stopped.
    """


class SkipPagination(Exception):
    """Raise within :class:`PaginationStategy` should be skipped, usually within
    :func:`~subgrounds.pagination.pagination.PaginationStrategy.__init___`.

    Causes :func:`~subgrounds.pagination.pagination.paginate` to be skipped.
    """


class PaginationStrategy(Protocol):
    def __init__(self, schema: SchemaMeta, document: Document) -> None:
        """Initializes the pagination strategy. If there is no need for pagination given
        the provided :class:`Document` ``document``, then the constructor should raise a
        :class:`SkipPagination` exception.

        Args:
            schema: The schema of the API against which ``document`` will be executed
            document: The query document
        """
        ...

    def step(
        self, page_data: dict[str, Any] | None = None
    ) -> tuple[Document, dict[str, Any]]:
        """Returns the new query document and its variables which will be executed to
        get the next page of data.

        If this is the first query made as part of the pagination strategy, then
          ``page_data`` will be ``None``.

        If pagination should be interupted (e.g.: if enough entities have been queried),
        then this method should raise a :class:`StopPagination` exception.

        Args:
            page_data: The previous query's response data. If this is the first query
              (i.e.: the first page of data), then it will be `None`. Defaults to None.

        Returns:
            tuple: A tuple `(doc, vars)` where `doc` is the query document that will be
              executed to fetch the next page of data and `vars` are the variables for
              that document.
        """
        ...


@dataclass
class LegacyStrategyArgGenerator:
    cursor: list[Cursor]
    active_idx: int = 0

    @dataclass
    class Cursor:
        page_node: PaginationNode

        inner: list[LegacyStrategyArgGenerator.Cursor]
        inner_idx: int = 0

        filter_value: Any = None
        queried_entities: int = 0
        stop: bool = False
        page_count: int = 0
        keys: set[str] = field(default_factory=set)

        def __init__(self, page_node: PaginationNode) -> None:
            self.page_node = page_node
            self.inner = list(page_node.inner | map(LegacyStrategyArgGenerator.Cursor))
            self.reset()

        @property
        def is_leaf(self):
            return len(self.inner) == 0

        def update(self, data: dict[str, Any]) -> None:
            """Moves `self` cursor forward according to previous response data `data`

            Args:
              data: Previous response data

            Raises:
              StopIteration: _description_
            """

            # Current node step
            index_field_data = list(
                extract_data(
                    [
                        *self.page_node.key_path,
                        *self.page_node.filter_field.split("__"),
                    ],
                    data,
                )
                | traverse
            )
            num_entities = len(index_field_data)
            filter_value = index_field_data[-1] if len(index_field_data) > 0 else None

            id_data = list(
                extract_data([*self.page_node.key_path, "id"], data) | traverse
            )
            for key in id_data:
                if key not in self.keys:
                    self.keys.add(key)

            self.page_count = self.page_count + 1
            self.queried_entities = len(self.keys)

            if filter_value:
                self.filter_value = filter_value

            if (
                (self.is_leaf and num_entities < PAGE_SIZE)
                or (not self.is_leaf and num_entities == 0)
                or (self.queried_entities == self.page_node.first_value)
            ):
                raise StopPagination

        def step(self, data: dict[str, Any]) -> None:
            """Updates either ``self`` cursor or inner state machine depending on
            whether the inner state machine has reached its limit

            Args:
              data: _description_
            """
            if self.is_leaf:
                self.update(data)
            else:
                try:
                    self.inner[self.inner_idx].step(data)
                except StopPagination:
                    if self.inner_idx < len(self.inner) - 1:
                        self.inner_idx = self.inner_idx + 1
                    else:
                        self.update(data)
                        self.inner_idx = 0

                    self.inner[self.inner_idx].reset()

        def first_arg_value(self) -> int:
            if self.is_leaf:
                return (
                    self.page_node.first_value - self.queried_entities
                    if self.page_node.first_value - self.queried_entities < PAGE_SIZE
                    else PAGE_SIZE
                )
            else:
                return 1

        def args(self) -> dict[str, Any]:
            """Returns the pagination arguments for the current state of the
             state machine

            Returns:
              _description_
            """

            args = {}
            args[f"first{self.page_node.node_idx}"] = self.first_arg_value()

            args[f"skip{self.page_node.node_idx}"] = (
                self.page_node.skip_value if self.page_count == 0 else 0
            )

            if self.filter_value is not None:
                args[f"lastOrderingValue{self.page_node.node_idx}"] = self.filter_value

            if self.is_leaf:
                return args
            else:
                inner_args = self.inner[self.inner_idx].args()
                return args | inner_args

        def reset(self):
            """Reset state machine"""

            self.inner_idx = 0
            self.filter_value = self.page_node.filter_value
            self.queried_entities = 0
            self.stop = False
            self.page_count = 0
            self.keys = set()

    def __init__(self, pagination_nodes: list[PaginationNode]) -> None:
        self.cursor = list(pagination_nodes | map(LegacyStrategyArgGenerator.Cursor))

    def step(self, page_data: Optional[dict[str, Any]] = None) -> dict[str, Any]:
        try:
            if page_data is not None:
                self.cursor[self.active_idx].step(page_data)

            return self.cursor[self.active_idx].args()
        except StopPagination:
            if self.active_idx < len(self.cursor) - 1:
                self.active_idx = self.active_idx + 1
                return self.cursor[self.active_idx].args()
            else:
                raise StopPagination


class LegacyStrategy:
    schema: SchemaMeta
    arg_generator: LegacyStrategyArgGenerator
    normalized_doc: Document

    def __init__(self, schema: SchemaMeta, document: Document) -> None:
        self.schema = schema

        pagination_nodes = generate_pagination_nodes(schema, document)

        if len(pagination_nodes) == 0:
            raise SkipPagination

        self.arg_generator = LegacyStrategyArgGenerator(pagination_nodes)
        self.normalized_doc = normalize(schema, document, pagination_nodes)

    def step(
        self, page_data: dict[str, Any] | None = None
    ) -> tuple[Document, dict[str, Any]]:
        args = self.arg_generator.step(page_data)
        trimmed_doc = prune_doc(self.normalized_doc, args)
        return trimmed_doc, args


@dataclass
class ShallowStrategyArgGenerator:
    @dataclass
    class Cursor:
        """Class used to generate the pagination variables for a given tree of
        ``PaginationNode`` objects.

        Attributes:
          page_node: The ``PaginationNode`` object which this cursor is iterating
            through
          inner: The cursors for nested ``PaginationNodes``, if any
          inner_idx: The index of the inner ``PaginationNode`` through which this cursor
            iterating
          filter_value: The previous page's index value used to query the next data page
            Depends on ``page_node.filter_field``, e.g.: if ``page_node.filter_field``
            is ``timestamp_gt``, then ``filter_value`` will be the highest timestamp
            the entities returned in the previous data page.
          queried_entities: Counter for the total number of queried entities
          stop: Flag indicating whether or not to stop the cursor
          page_count: Counter keeping track of the total number data pages queried
          keys: Set of keys of all queried entities to avoid duplicates

        # TODO: Add GreedyStrategy doc
        """

        page_node: PaginationNode

        inner: list[ShallowStrategyArgGenerator.Cursor]
        inner_idx: int = 0

        filter_value: Any = None
        queried_entities: int = 0
        page_count: int = 0

        @staticmethod
        def from_pagination_node(
            page_node: PaginationNode,
        ) -> ShallowStrategyArgGenerator.Cursor:
            return ShallowStrategyArgGenerator.Cursor(
                page_node=page_node,
                filter_value=page_node.filter_value,
                inner=list(
                    page_node.inner
                    | map(ShallowStrategyArgGenerator.Cursor.from_pagination_node)
                ),
            )

        @property
        def is_leaf(self) -> bool:
            return len(self.inner) == 0

        @property
        def active_cursor(self) -> Optional[ShallowStrategyArgGenerator.Cursor]:
            if len(self.inner) > 0:
                return self.inner[self.inner_idx]
            else:
                return None

        def iter(
            self, only_active: bool = False
        ) -> Iterator[ShallowStrategyArgGenerator.Cursor]:
            yield self
            if only_active and len(self.inner) > 0:
                yield from self.inner[self.inner_idx].iter(only_active=True)
            else:
                for cursor in self.inner:
                    yield from cursor.iter()

        def mapi(
            self,
            map_f: Callable[
                [int, ShallowStrategyArgGenerator.Cursor],
                ShallowStrategyArgGenerator.Cursor,
            ],
            priority: Literal["self"] | Literal["children"] = "self",
            counter: count[int] | None = None,
        ) -> ShallowStrategyArgGenerator.Cursor:
            if counter is None:
                counter = count()

            match priority:
                case "self":
                    new_cursor = map_f(next(counter), self)

                    return ShallowStrategyArgGenerator.Cursor(
                        page_node=new_cursor.page_node,
                        inner=list(
                            new_cursor.inner
                            | map(
                                partial(
                                    ShallowStrategyArgGenerator.Cursor.mapi,
                                    map_f=map_f,
                                    counter=counter,
                                )
                            )
                        ),
                        inner_idx=new_cursor.inner_idx,
                        filter_value=new_cursor.filter_value,
                        queried_entities=new_cursor.queried_entities,
                        page_count=new_cursor.page_count,
                    )

                case "children":
                    i = next(counter)

                    new_cursor = ShallowStrategyArgGenerator.Cursor(
                        page_node=self.page_node,
                        inner=list(
                            self.inner
                            | map(
                                partial(
                                    ShallowStrategyArgGenerator.Cursor.mapi,
                                    map_f=map_f,
                                    counter=counter,
                                )
                            )
                        ),
                        inner_idx=self.inner_idx,
                        filter_value=self.filter_value,
                        queried_entities=self.queried_entities,
                        page_count=self.page_count,
                    )

                    return map_f(i, new_cursor)

                case _:
                    raise Exception(f"Cursor.mapi: invalid priority {priority}")

    cursor: list[Cursor]

    def __init__(self, pagination_nodes: list[PaginationNode]) -> None:
        self.cursor = list(
            pagination_nodes
            | map(ShallowStrategyArgGenerator.Cursor.from_pagination_node)
        )

    def iter_cursors(self) -> Iterator[Cursor]:
        for cur in self.cursor:
            yield from cur.iter()

    @staticmethod
    def update_cursor(cursor: Cursor, data: dict[str, Any]) -> Cursor:
        index_field_data = list(
            extract_data(
                [*cursor.page_node.key_path, cursor.page_node.filter_field], data
            )
            | traverse
        )
        num_entities = len(index_field_data)
        filter_value = index_field_data[-1] if len(index_field_data) > 0 else None

        new_cursor = ShallowStrategyArgGenerator.Cursor(
            page_node=cursor.page_node,
            inner=cursor.inner,
            inner_idx=0,
            filter_value=filter_value,
            queried_entities=cursor.queried_entities + num_entities,
            page_count=cursor.page_count + 1,
        )

        if (
            new_cursor.queried_entities >= new_cursor.page_node.first_value
            or num_entities < PAGE_SIZE
        ):
            raise StopPagination

        return new_cursor

    def step(self, page_data: Optional[dict[str, Any]] = None) -> dict[str, Any]:
        if page_data is not None:
            self.cursor = list(
                self.cursor
                | map(
                    partial(
                        ShallowStrategyArgGenerator.Cursor.mapi,
                        map_f=lambda i, cursor: (
                            ShallowStrategyArgGenerator.update_cursor(cursor, page_data)
                            if i == 0
                            else cursor
                        ),
                    )
                )
            )

        args = {}
        for cur in self.iter_cursors():
            args[f"first{cur.page_node.node_idx}"] = (
                cur.page_node.first_value - cur.queried_entities
                if cur.page_node.first_value - cur.queried_entities < PAGE_SIZE
                else PAGE_SIZE
            )

            args[f"skip{cur.page_node.node_idx}"] = (
                cur.page_node.skip_value if cur.page_count == 0 else 0
            )

            if cur.filter_value is not None:
                args[f"lastOrderingValue{cur.page_node.node_idx}"] = cur.filter_value

        return args


class ShallowStrategy:
    schema: SchemaMeta
    arg_generator: ShallowStrategyArgGenerator
    normalized_doc: Document

    def __init__(self, schema: SchemaMeta, document: Document) -> None:
        self.schema = schema

        pagination_nodes = generate_pagination_nodes(schema, document)

        if len(pagination_nodes) == 0:
            raise SkipPagination

        self.arg_generator = ShallowStrategyArgGenerator(pagination_nodes)
        self.normalized_doc = normalize(schema, document, pagination_nodes)

    def step(
        self, page_data: dict[str, Any] | None = None
    ) -> tuple[Document, dict[str, Any]]:
        args = self.arg_generator.step(page_data)
        trimmed_doc = prune_doc(self.normalized_doc, args)
        return (trimmed_doc, args)


class SkipStrategy:
    """This strategy always raises a :class:`~subgrounds.pagination.SkipPagination`
      when constructed or if *somehow* `step` is called.

    This is the default if `None` is passed to the strategy (default for non-subgraphs).
    """

    def __init__(self, schema: SchemaMeta, document: Document) -> None:
        raise SkipPagination

    def step(
        self, page_data: Optional[dict[str, Any]] = None
    ) -> tuple[Document, dict[str, Any]]:
        raise StopPagination


def normalize_strategy(
    strategy: type[PaginationStrategy] | None = LegacyStrategy,
    is_subgraph: bool = True,
):
    if strategy is not None and is_subgraph:
        return strategy

    return SkipStrategy



from subgrounds.schema import TypeRef

from .base import DocumentTransform, RequestTransform
from .transforms import TypeTransform

DEFAULT_GLOBAL_TRANSFORMS: list[RequestTransform] = []

DEFAULT_SUBGRAPH_TRANSFORMS: list[DocumentTransform] = [
    TypeTransform(
        TypeRef.Named(name="BigDecimal", kind="SCALAR"),
        lambda bigdecimal: float(bigdecimal),
    ),
    TypeTransform(
        TypeRef.Named(name="BigInt", kind="SCALAR"),
        lambda bigint: int(bigint),
    ),
]



""" Top-level Subgrounds subpackage

This subpackage implements the top-level API that most developers will be using when
 querying subgraphs with Subgrounds. `Subgrounds` and `AsyncSubgrounds` contains fully
 implemented subgraph querying SDKs for accessing subgraph data (flattened or nested)
 effectively with transformation and pagination bundled in.

`SubgroundsBase` provides an extended interface to build custom clients on-top allowing
 for intricate customization on the buisness logic subgrounds implements.
"""

from .async_ import AsyncSubgrounds
from .base import SubgroundsBase
from .sync import Subgrounds

__all__ = ["AsyncSubgrounds", "Subgrounds", "SubgroundsBase"]



""" Helper functions and classes used by Subgrounds' own pagination strategies.
"""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from functools import reduce
from itertools import count
from typing import Any, cast

from pipe import map, reverse, skip, traverse

from subgrounds.query import (
    Argument,
    Document,
    InputValue,
    Selection,
    VariableDefinition,
)
from subgrounds.schema import SchemaMeta, TypeMeta, TypeRef
from subgrounds.utils import accumulate

from .utils import merge_input_value_object_metas

DEFAULT_NUM_ENTITIES = 100


@dataclass(frozen=True)
class PaginationNode:
    """Class representing the pagination config for a single GraphQL list field.

    Attributes:
      node_idx: Index of PaginationNode, used to label pagination arguments for
        this node.
      filter_field : Name of the node's filter field, e.g.: if ``filter_name`` is
        ``timestamp_gt``, then :attr:`filter_field` is ``timestamp``
      first_value: Initial value of the ``first`` argument
      skip_value: Initial value of the ``skip`` argument
      filter_value: Initial value of the filter argument
        (i.e.: ``where: {filter: FILTER_VALUE}``)
      filter_value_type: Type of the filter value
      key_path: Location in the list field to which this pagination node refers to in
        the initial query
      inner: Nested pagination nodes (if any).
    """

    node_idx: int
    filter_field: str

    first_value: int
    skip_value: int
    filter_value: Any
    filter_value_type: TypeRef.T

    key_path: list[str]
    inner: list[PaginationNode] = field(default_factory=list)

    def get_vardefs(self) -> list[VariableDefinition]:
        """Returns a list of variable definitions corresponding to this pagination
        node's pagination arguments as well as the variable definitions related
        to any nested pagination nodes.

        Returns:
          _description_
        """
        vardefs = [
            VariableDefinition(
                f"first{self.node_idx}", TypeRef.Named(name="Int", kind="SCALAR")
            ),
            VariableDefinition(
                f"skip{self.node_idx}", TypeRef.Named(name="Int", kind="SCALAR")
            ),
            VariableDefinition(
                f"lastOrderingValue{self.node_idx}", self.filter_value_type
            ),
        ]

        nested_vardefs = list(self.inner | map(PaginationNode.get_vardefs) | traverse)

        return nested_vardefs + vardefs


PAGINATION_ARGS: set[str] = {"first", "skip", "where", "orderBy", "orderDirection"}


def is_pagination_node(schema: SchemaMeta, selection: Selection) -> bool:
    return (
        selection.fmeta.type_.is_list
        and schema.type_of_typeref(selection.fmeta.type_).is_object
    )


def get_orderBy_value(selection: Selection) -> str:
    order_by_val = selection.find_args(lambda arg: arg.name == "orderBy", recurse=False)
    if order_by_val is None:
        return "id"

    return order_by_val.value.value


def get_orderDirection_value(selection: Selection) -> str:
    order_direction_arg = selection.find_args(
        lambda arg: arg.name == "orderDirection", recurse=False
    )
    if order_direction_arg is None:
        return "asc"

    return order_direction_arg.value.value


def get_filtering_args(selection: Selection) -> list[str]:
    orderBy_val = get_orderBy_value(selection)
    orderDirection_val = get_orderDirection_value(selection)

    *names, last = orderBy_val.split("__")

    return [*names, "{}_{}".format(last, "gt" if orderDirection_val == "asc" else "lt")]


def get_filtering_value(selection: Selection) -> Any | None:
    def recurse_filtering_values(
        filtering_args: list[str],
        obj: InputValue.Object,
    ) -> Any | None:
        if not filtering_args:
            return None

        if (filtering_arg := filtering_args.pop(0)) not in obj.value:
            return None

        match obj.value[filtering_arg]:
            case InputValue.Object() as obj:
                return recurse_filtering_values(filtering_args, obj)
            case value:
                return value.value

    where_arg = selection.find_args(lambda arg: arg.name == "where", recurse=False)
    if where_arg is None:
        return None

    return recurse_filtering_values(
        get_filtering_args(selection), cast(InputValue.Object, where_arg.value)
    )


def generate_pagination_nodes(
    schema: SchemaMeta, document: Document
) -> list[PaginationNode]:
    counter = count()

    def fold_f(
        current: Selection, parents: list[Selection], children: list[PaginationNode]
    ) -> PaginationNode | list[PaginationNode]:
        if not is_pagination_node(schema, current):
            return children
        idx = next(counter)

        orderBy_val = get_orderBy_value(current)
        filtering_args = get_filtering_args(current)

        t: TypeRef.T = current.fmeta.type_of_arg("where")
        where_arg_type: TypeMeta.InputObjectMeta = schema.type_of_typeref(t)
        filtering_arg_type: TypeRef.T = schema.type_of_input_object_meta(
            where_arg_type, filtering_args
        )

        return PaginationNode(
            node_idx=idx,
            filter_field=orderBy_val,
            first_value=(
                current.find_args(
                    lambda arg: arg.name == "first", recurse=False
                ).value.value
                if current.exists_args(lambda arg: arg.name == "first", recurse=False)
                else DEFAULT_NUM_ENTITIES
            ),
            skip_value=(
                current.find_args(
                    lambda arg: arg.name == "skip", recurse=False
                ).value.value
                if current.exists_args(lambda arg: arg.name == "skip", recurse=False)
                else 0
            ),
            filter_value=get_filtering_value(current),
            filter_value_type=filtering_arg_type,
            key_path=[select.key for select in [*parents, current]],
            inner=children,
        )

    return list(document.query.fold(fold_f) | traverse)


def normalize(
    schema: SchemaMeta,
    document: Document,
    pagination_nodes: list[PaginationNode],
):
    """Inject various graphql components to "normalize" the query for pagination.

    When we paginate a query, we inject custom filtering based on the order by values.
    We also add GraphQL variables so that `PaginationStrategy` only need to change those
     to perform pagination.

    The main process for normalization begins by recursively adjusting `Selection` nodes
     within the Query tree. We only apply the following steps if the node needs to be
     paginated.

    > Note, these steps always check the current selection and will merge new values
     and selections onto whats currently there.

    1. Ensure `id` is on the `Selection`
    2. Replace `first` argument value by `$firstX`
    3. Replace `skip` argument value by `$skipX`
    4. With the `orderBy` (default being `id`), generate where filtering arguments
      a) These are used to filter out values when paginating
    5. Set `where` filtering values (deep union / merge)
    """

    counter = count()

    def map_f(current: Selection) -> Selection:
        if not is_pagination_node(schema, current):
            return current

        idx = next(counter)

        where_value: dict[str, Any] = cast(
            dict,
            current.find_args(
                lambda arg: arg.name == "where", recurse=False
            ).value.value  # type: ignore
            if current.exists_args(lambda arg: arg.name == "where", recurse=False)
            else {},
        )

        orderBy_value = get_orderBy_value(current)

        # Build out a nested dictionary of `InputValue.Object` containing the
        #  `lastOrderingValue{idx}` used for pagination. We do this by iterating through
        #  the args backwards starting from the innermost and nesting them in dicts.
        # We need to add a trailing `_` to outer arguments for `InputValue.Object` for
        #  where clauses to work properly.
        *args, innermost_arg = get_filtering_args(current)
        where_filtering_args: InputValue.Object = reduce(
            lambda inner, outer_arg: InputValue.Object({f"{outer_arg}_": inner}),
            args | reverse,
            InputValue.Object(
                {innermost_arg: InputValue.Variable(f"lastOrderingValue{idx}")}
            ),
        )

        pagination_args = [
            Argument(name="first", value=InputValue.Variable(f"first{idx}")),
            Argument(name="skip", value=InputValue.Variable(f"skip{idx}")),
            Argument(name="orderBy", value=InputValue.Enum(orderBy_value)),
            Argument(
                name="orderDirection",
                value=InputValue.Enum(get_orderDirection_value(current)),
            ),
            Argument(
                name="where",
                value=InputValue.Object(
                    merge_input_value_object_metas(
                        where_filtering_args.value, where_value
                    )
                ),
            ),
        ]

        # This ensures `id` always exists
        current = current.add(
            Selection(
                fmeta=TypeMeta.FieldMeta(
                    name="id",
                    description="",
                    args=[],
                    type=TypeRef.Named(name="String", kind="SCALAR"),
                )
            )
        )

        # Using nested orderBy values (tabulated by "__"), gather the type of the field
        #  from the schema.
        orderBy_values = orderBy_value.split("__")

        # Note, we skip the first type since that is the type of the current `Selection`
        orderBy_types = (
            orderBy_values
            | accumulate(
                func=lambda curr, val: (
                    schema.type_of_typeref(curr).type_of_field(val)
                ),
                initial=schema.type_of_typeref(current.fmeta.type_),
            )
            | skip(1)
        )

        # Generate a `Selection` tree to add to the current selection.
        # This tree is generated from the orderBy values and types as nested `FieldMeta`
        #  constructed through `reduce` by placing singular `Selection` objects in the
        #  `selection` field on the `Selection`.
        current = current.add(
            reduce(
                lambda current, new: replace(current, selection=[new]),
                zip(orderBy_values, orderBy_types)
                | map(
                    lambda tup: Selection(
                        fmeta=TypeMeta.FieldMeta(
                            name=tup[0],
                            description="",
                            args=[],
                            type=tup[1],
                        ),
                    )
                ),
            )
        )

        # add the other arguments that aren't used for pagination
        pagination_args += current.find_all_args(
            lambda arg: arg.name not in PAGINATION_ARGS, recurse=False
        )

        return replace(current, arguments=pagination_args)

    query = document.query.map(map_f, priority="children")
    vardefs = list(pagination_nodes | map(PaginationNode.get_vardefs) | traverse)

    # TODO: normalize document fragments
    return replace(document, query=query.add_vardefs(vardefs))


def prune_doc(document: Document, args: dict[str, Any]) -> Document:
    def prune_where_arg(where_arg: Argument) -> Argument:
        input_val: InputValue.Object = where_arg.value
        return Argument(
            name=where_arg.name,
            value=InputValue.Object(
                {
                    name: val
                    for name, val in input_val.value.items()
                    if not val.is_variable or (val.is_variable and val.name in args)
                }
            ),
        )

    return (
        document.map_args(
            lambda arg: prune_where_arg(arg) if arg.name == "where" else arg
        )
        .filter_args(
            lambda arg: (
                arg.find_var(lambda var: "lastOrderingValue" in var.name).name in args
                if arg.exists_vars(lambda var: "lastOrderingValue" in var.name)
                else True
            )
        )
        .prune_undefined(args)
    )



from __future__ import annotations

import logging
import operator
import warnings
from dataclasses import dataclass
from datetime import datetime
from functools import partial, reduce
from hashlib import blake2b
from typing import TYPE_CHECKING, Any, Callable, ClassVar

from pipe import map, traverse
from typing_extensions import Self  # 3.10 support

from subgrounds.query import Query, Selection, arguments_of_field_args
from subgrounds.schema import SchemaMeta, TypeMeta, TypeRef
from subgrounds.utils import extract_data

from .filter import Filter

if TYPE_CHECKING:
    from .subgraph import Subgraph

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")

FPATH_DEPTH_LIMIT: int = 4


def typeref_of_binary_op(
    op: str, t1: TypeRef.T, t2: int | float | str | bool | FieldPath | SyntheticField
):
    def f_typeref(t1, t2):
        match (op, TypeRef.root_type_name(t1), TypeRef.root_type_name(t2)):
            case ("add", "String" | "Bytes", "String" | "Bytes"):
                return TypeRef.Named(name="String", kind="SCALAR")

            case (
                "add" | "sub" | "mul" | "div" | "pow" | "mod",
                "BigInt" | "Int",
                "BigInt" | "Int",
            ):
                return TypeRef.Named(name="Int", kind="SCALAR")
            case (
                "add" | "sub" | "mul" | "div" | "pow",
                "BigInt" | "Int",
                "BigDecimal" | "Float",
            ):
                return TypeRef.Named(name="Float", kind="SCALAR")
            case (
                "add" | "sub" | "mul" | "div" | "pow",
                "BigDecimal" | "Float",
                "BigInt" | "Int" | "BigDecimal" | "Float",
            ):
                return TypeRef.Named(name="Float", kind="SCALAR")

            case _ as args:
                raise Exception(
                    f"typeref_of_binary_op: f_typeref: unhandled arguments {args}"
                )

    def f_const(t1, const):
        match (op, TypeRef.root_type_name(t1), const):
            case ("add", "String" | "Bytes", str()):
                return TypeRef.Named(name="String", kind="SCALAR")

            case (
                "add" | "sub" | "mul" | "div" | "pow" | "mod",
                "BigInt" | "Int",
                int(),
            ):
                return TypeRef.Named(name="Int", kind="SCALAR")
            case ("add" | "sub" | "mul" | "div" | "pow", "BigInt" | "Int", float()):
                return TypeRef.Named(name="Float", kind="SCALAR")
            case (
                "add" | "sub" | "mul" | "div" | "pow",
                "BigDecimal" | "Float",
                int() | float(),
            ):
                return TypeRef.Named(name="Float", kind="SCALAR")

            case _ as args:
                raise Exception(
                    f"typeref_of_binary_op: f_typeref: unhandled arguments {args}"
                )

    match t2:
        case int() | float() | str() | bool() as constant:
            return f_const(t1, constant)
        case FieldPath() | SyntheticField() as field:
            return f_typeref(t1, field._type)


def type_ref_of_unary_op(op: str, t: TypeRef.T):
    match (op, TypeRef.root_type_name(t)):
        case ("abs", "BigInt" | "Int"):
            return TypeRef.Named(name="Int", kind="SCALAR")
        case ("abs", "BigDecimal" | "Float"):
            return TypeRef.Named(name="Float", kind="SCALAR")

        case ("neg", "BigInt" | "Int"):
            return TypeRef.Named(name="Int", kind="SCALAR")
        case ("neg", "BigDecimal" | "Float"):
            return TypeRef.Named(name="Float", kind="SCALAR")

        case _ as args:
            raise Exception(
                f"typeref_of_binary_op: f_typeref: unhandled arguments {args}"
            )


class FieldOperatorMixin:
    _subgraph: Subgraph
    _type: TypeRef.T

    def __add__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            operator.add, typeref_of_binary_op("add", self._type, other), [self, other]
        )

    def __radd__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.add(y, x),
            typeref_of_binary_op("add", self._type, other),
            [self, other],
        )

    def __sub__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            operator.sub, typeref_of_binary_op("sub", self._type, other), [self, other]
        )

    def __rsub__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.sub(y, x),
            typeref_of_binary_op("sub", self._type, other),
            [self, other],
        )

    def __mul__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            operator.mul, typeref_of_binary_op("mul", self._type, other), [self, other]
        )

    def __rmul__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.mul(y, x),
            typeref_of_binary_op("mul", self._type, other),
            [self, other],
        )

    def __truediv__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            operator.truediv,
            typeref_of_binary_op("div", self._type, other),
            [self, other],
        )

    def __rtruediv__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.truediv(y, x),
            typeref_of_binary_op("div", self._type, other),
            [self, other],
        )

    def __floordiv__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            operator.floordiv,
            typeref_of_binary_op("div", self._type, other),
            [self, other],
        )

    def __rfloordiv__(self, other: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.floordiv(y, x),
            typeref_of_binary_op("div", self._type, other),
            [self, other],
        )

    def __pow__(self, rhs: Any) -> SyntheticField:
        return SyntheticField(
            operator.pow, typeref_of_binary_op("pow", self._type, rhs), [self, rhs]
        )

    def __rpow__(self, lhs: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.pow(y, x),
            typeref_of_binary_op("pow", self._type, lhs),
            [self, lhs],
        )

    def __mod__(self, rhs: Any) -> SyntheticField:
        return SyntheticField(
            operator.mod, typeref_of_binary_op("mod", self._type, rhs), [self, rhs]
        )

    def __rmod__(self, lhs: Any) -> SyntheticField:
        return SyntheticField(
            lambda x, y: operator.mod(y, x),
            typeref_of_binary_op("mod", self._type, lhs),
            [self, lhs],
        )

    def __neg__(self) -> SyntheticField:
        return SyntheticField(
            operator.neg, type_ref_of_unary_op("neg", self._type), self
        )

    def __abs__(self) -> SyntheticField:
        return SyntheticField(
            operator.abs, type_ref_of_unary_op("abs", self._type), self
        )


def fieldpaths_of_object(
    subgraph: Subgraph, object_: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta
):
    """Returns generator of FieldPath objects that selects all non-list fields of
    GraphQL Object of Interface :attr:`object_`.

    Args:
      schema: _description_
      object_: _description_

    Yields:
      _description_
    """
    for fmeta in object_.fields:
        if not fmeta.type_.is_list and len(fmeta.arguments) == 0:
            match subgraph._schema.type_of_typeref(fmeta.type_):
                case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta():
                    yield subgraph.__getattribute__(object_.name).__getattribute__(
                        fmeta.name
                    ).id

                case _:
                    yield subgraph.__getattribute__(object_.name).__getattribute__(
                        fmeta.name
                    )


@dataclass
class FieldPath(FieldOperatorMixin):
    _subgraph: Subgraph
    _root_type: TypeRef.T
    _type: TypeRef.T
    _path: list[tuple[dict[str, Any] | None, TypeMeta.FieldMeta]]

    # Purely for testing
    __test_mode: ClassVar[bool] = False

    def __init__(
        self,
        subgraph: Subgraph,
        root_type: TypeRef.T,
        type_: TypeRef.T,
        path: list[tuple[dict[str, Any] | None, TypeMeta.FieldMeta]],
    ) -> None:
        self._subgraph = subgraph
        self._root_type = root_type
        self._type = type_
        self._path = path

        # Add fields as attributes if leaf is object
        match self._subgraph._schema.type_of(self._leaf):
            case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta() as type_ if len(
                self._path
            ) < FPATH_DEPTH_LIMIT:
                # We generate fieldpaths up to depth 8
                for fmeta in type_.fields:
                    path = self._path.copy()
                    path.append((None, fmeta))

                    super().__setattr__(
                        fmeta.name,
                        FieldPath(
                            subgraph=self._subgraph,
                            root_type=self._root_type,
                            type_=fmeta.type_,
                            path=path,
                        ),
                    )

            case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta() as type_ if len(
                self._path
            ) == FPATH_DEPTH_LIMIT:
                for fmeta in type_.fields:
                    # NOTE: We set the attribute to None on purpose since we want code
                    # completion to work while avoiding infinite loops caused by cycles
                    # in the GraphQL schema. The attribute itself will be initialized
                    # on __getattribute_
                    super().__setattr__(fmeta.name, None)

            case _:
                pass

    @property
    def _schema(self) -> SchemaMeta:
        return self._subgraph._schema

    @property
    def _root(self) -> TypeMeta.FieldMeta:
        """Returns the type information of the root field of the current
        :class:`FieldPath`

        Returns:
          Type information of the root field of the current :class:`FieldPath`
        """
        return self._path[0][1]

    @property
    def _leaf(self) -> TypeMeta.FieldMeta:
        """Returns the type information of the leaf field of the current
        :class:`FieldPath`

        Returns:
          Type information of the leaf field of the current :class:`FieldPath`
        """
        return self._path[-1][1]

    @staticmethod
    def _hash(msg: str) -> str:
        h = blake2b(digest_size=8)
        h.update(msg.encode("UTF-8"))
        return "x" + h.hexdigest()

    @staticmethod
    def _merge(fpaths: list[FieldPath]) -> list[Selection]:
        """Returns a Selection tree containing all selection paths in `fpaths`.
        Note: Assumes that all fieldpaths in `fpaths` belong to the same subgraph

        Args:
          fpaths: _description_

        Returns:
          _description_
        """

        query = reduce(Query.add, fpaths | map(FieldPath._selection), Query())
        return query.selection

    def _name_path(self, use_aliases: bool = False) -> list[str]:
        """Returns a list of strings correspoding to the names of all fields
        selected in the current :class:`FieldPath`. If :attr:`use_aliases` is True,
        then if a field has an automatically generated alias, the alias will be
        returned.

        Args:
          use_aliases: Flag indicating wether of not to use the fields' automatically
            generated alias (if present). Defaults to False.

        Returns:
          List of field names selected in the current :class:`FieldPath`
        """

        def gen_alias(ele: tuple[dict[str, Any] | None, TypeMeta.FieldMeta]) -> str:
            if ele[0] != {} and ele[0] is not None:
                return FieldPath._hash(ele[1].name + str(ele[0]))
            else:
                return ele[1].name

        return list(
            self._path | map(lambda ele: gen_alias(ele) if use_aliases else ele[1].name)
        )

    def _name(self, use_aliases: bool = False) -> str:
        """Generates the name of the current :class:`FieldPath` using the names of
        the fields it selects. If :attr:`use_aliases` is True, then if a field has
        an automatically generated alias, the alias will be used.

        Args:
          use_aliases: Flag indicating wether of not to use the fields' automatically
            generated alias (if present). Defaults to False.

        Returns:
          The generated name of the current :class:`FieldPath`.
        """

        return "_".join(self._name_path(use_aliases=use_aliases))

    def _auto_select(self) -> FieldPath | list[FieldPath]:
        match self._subgraph._schema.type_of_typeref(self._leaf.type_):
            case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta() as obj:
                return list(
                    fieldpaths_of_object(self._subgraph, obj)
                    | map(partial(FieldPath._extend, self))
                )

            case _:
                return self

    def _extract_data(
        self, data: dict[str, Any] | list[dict[str, Any]]
    ) -> list[Any] | Any:
        """Extract the data corresponding to the current :class:`FieldPath` from
        the dictionary :attr:`data`.

        Args:
          data: Data dictionary that contains the data corresponding to the current
            :class:`FieldPath`.

        Returns:
          Data corresponding to the current :class:`FieldPath`.
        """
        return extract_data(self._name_path(use_aliases=True), data)

    def _selection(self) -> Selection | list[Selection]:
        """Returns a selection or list of selections corresponding to the current
        :class:`FieldPath`.

        Returns:
          Selection | list[Selection]: _description_
        """

        def f(
            path: list[tuple[dict[str, Any] | None, TypeMeta.FieldMeta]]
        ) -> list[Selection]:
            match path:
                case [
                    (args, TypeMeta.FieldMeta() as fmeta),
                    *rest,
                ] if args == {} or args is None:
                    return [Selection(fmeta, selection=f(rest))]

                case [(args, TypeMeta.FieldMeta() as fmeta), *rest]:
                    return [
                        Selection(
                            fmeta,
                            # TODO: Revisit this
                            alias=FieldPath._hash(fmeta.name + str(args)),
                            arguments=arguments_of_field_args(
                                self._subgraph._schema, fmeta, args
                            ),
                            selection=f(rest),
                        )
                    ]

                case []:
                    return []

            assert False  # Suppress mypy missing return statement warning

        return f(self._path)[0]

    def _set_arguments(
        self, args: dict[str, Any], selection: list[FieldPath] = []
    ) -> FieldPath | list[FieldPath]:
        """Set the arguments to the leaf of the current :class:`FieldPath`. The
        method returns the :attr:`self`.

        Args:
          args: _description_
          selection: _description_. Defaults to [].

        Returns:
          _description_
        """

        def fmt_arg(name, raw_arg):
            match (name, raw_arg):
                case ("where", [Filter(), *_] as filters):
                    return Filter.to_dict(filters)
                case ("orderBy", FieldPath() as fpath):
                    if paths := fpath._name_path():
                        return "__".join(paths)
                    raise Exception(
                        f"Cannot use empty paths as orderBy argument {fpath}"
                    )
                case _:
                    return raw_arg

        match self._leaf:
            case TypeMeta.FieldMeta():
                args = {key: fmt_arg(key, val) for key, val in args.items()}
                self._path[-1] = (args, self._path[-1][1])
                if len(selection) > 0:
                    return list(selection | map(partial(FieldPath._extend, self)))
                else:
                    return self
            case _:
                raise TypeError(f"Unexpected type for FieldPath {self}")

    def _select(self, name: str) -> FieldPath:
        """Returns a new FieldPath corresponding to the FieldPath `self` extended with
        an additional selection on the field named `name`.

        Args:
          name: The name of the field to expand on the leaf of `fpath`

        Raises:
          TypeError: [description]
          TypeError: [description]
          TypeError: [description]

        Returns:
          A new FieldPath containing `fpath` extended with the field named `name`
        """
        match self._schema.type_of_typeref(self._type):
            # If the FieldPath fpath
            case TypeMeta.EnumMeta() | TypeMeta.ScalarMeta():
                raise TypeError(
                    f"FieldPath: path {self} ends with a scalar field!"
                    f" cannot select field {name}"
                )

            case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta() as obj:
                field = obj.field(name)

                match self._schema.type_of_typeref(field.type_):
                    case (
                        TypeMeta.ObjectMeta()
                        | TypeMeta.InterfaceMeta()
                        | TypeMeta.EnumMeta()
                        | TypeMeta.ScalarMeta()
                    ):
                        # Copy current path and append newly selected field
                        path = self._path.copy()
                        path.append((None, field))

                        # Return new FieldPath
                        return FieldPath(
                            subgraph=self._subgraph,
                            root_type=self._root_type,
                            type_=field.type_,
                            path=path,
                        )
                    case _:
                        raise TypeError(
                            f"FieldPath: field {name} is not a valid field for object"
                            f" {self._type.name} at path {self}"
                        )

            case _:
                raise TypeError(
                    f"FieldPath: Unexpected type {self._type.name}"
                    f" when selection {name} on {self}"
                )

    def _extend(self, ext: FieldPath) -> FieldPath:
        """Extends the current :class:`FieldPath` with the :class:`FieldPath`
        :attr:`ext`. :attr:`ext` must start where the current :class:`FieldPath` ends.

        Args:
          ext: The :class:`FieldPath` representing the extension

        Raises:
          TypeError: [description]
          TypeError: [description]
          TypeError: [description]

        Returns:
          A new :class:`FieldPath` containing the initial current :class:`FieldPath`
            extended with :attr:`ext`
        """
        match self._leaf:
            case TypeMeta.FieldMeta() as fmeta:
                match self._schema.type_of_typeref(fmeta.type_):
                    case TypeMeta.ObjectMeta(name=name) | TypeMeta.InterfaceMeta(
                        name=name
                    ):
                        if name == ext._root_type.name:
                            return FieldPath(
                                subgraph=self._subgraph,
                                root_type=self._root_type,
                                type_=ext._type,
                                path=self._path + ext._path,
                            )
                        else:
                            raise TypeError(
                                f"extend: FieldPath {ext} does not start at the"
                                f" same type from where FieldPath {self} ends"
                            )
                    case _:
                        raise TypeError(f"extend: FieldPath {self} is not object field")
            case _:
                raise TypeError(f"extend: FieldPath {self} is not an object field")

    # ================================================================
    # Overloaded magic functions
    # ================================================================
    # When setting arguments
    def __call__(self, **kwargs: Any) -> Any:
        """Sets field arguments and expand subfields. The updated FieldPath is returned.

        Example:
        >>> aaveV2 = sg.load_subgraph(
        ...     "https://api.thegraph.com/subgraphs/name/aave/protocol-v2"
        ... )
        >>> query = aaveV2.Query.borrows(
        ...   first=10,
        ...   order_by=aaveV2.Borrow.timestamp,
        ...   order_direction="desc",
        ...   selection=[
        ...     aaveV2.Borrow.id,
        ...     aaveV2.Borrow.timestamp,
        ...     aaveV2.Borrow.amount
        ...   ]
        ... )

        Returns:
          The updated field path if :attr:`selection` is not specified, or a list of
            fieldpaths when :attr:`selection` is specified.
        """
        selection = kwargs.pop("selection", [])
        return self._set_arguments(kwargs, selection)

    # Field selection
    def __getattribute__(self, __name: str) -> Any:
        # Small hack to get code completion to work while allowing updates to FieldPath
        # (i.e.: setting arguments)
        try:
            match super().__getattribute__(__name):
                case FieldPath() | SyntheticField() | None:
                    return self._select(__name)
                case value:
                    return value
        except AttributeError:
            return self._select(__name)

    # Filtering
    def __eq__(self, value: FieldPath | Any) -> Filter | bool:
        if FieldPath.__test_mode:
            # Purely used for testing so that assertEqual works
            return (
                self._subgraph == value._subgraph
                and self._type == value._type
                and self._path == value._path
            )
        else:
            return Filter.mk_filter(self, Filter.Operator.EQ, value)

    def __ne__(self, value: Any) -> Filter:
        return Filter.mk_filter(self, Filter.Operator.NEQ, value)

    def __lt__(self, value: Any) -> Filter:
        return Filter.mk_filter(self, Filter.Operator.LT, value)

    def __gt__(self, value: Any) -> Filter:
        return Filter.mk_filter(self, Filter.Operator.GT, value)

    def __le__(self, value: Any) -> Filter:
        return Filter.mk_filter(self, Filter.Operator.LTE, value)

    def __ge__(self, value: Any) -> Filter:
        return Filter.mk_filter(self, Filter.Operator.GTE, value)

    # Utility
    def __str__(self) -> str:
        return ".".join(self._path | map(lambda ele: ele[1].name))

    def __repr__(self) -> str:
        vars = f"{self._subgraph._url}, {self._root_type.name}, {self._name_path()}"
        return f"FieldPath({vars})"


@dataclass
class SyntheticField(FieldOperatorMixin):
    STRING: ClassVar[TypeRef.Named] = TypeRef.Named(name="String", kind="SCALAR")
    INT: ClassVar[TypeRef.Named] = TypeRef.Named(name="Int", kind="SCALAR")
    FLOAT: ClassVar[TypeRef.Named] = TypeRef.Named(name="Float", kind="SCALAR")
    BOOL: ClassVar[TypeRef.Named] = TypeRef.Named(name="Boolean", kind="SCALAR")

    _counter: ClassVar[int] = 0

    _f: Callable
    _type: TypeRef.T
    _default: Any
    _deps: list[FieldPath]

    def __init__(
        self,
        f: Callable,
        type_: TypeRef.T,
        deps: list[FieldPath | SyntheticField] | FieldPath | SyntheticField,
        default: Any = None,
    ) -> None:
        deps = list([deps] | traverse)

        def mk_deps(
            deps: list[FieldPath | SyntheticField],
            f: Callable,
            acc: list[tuple[Callable | None, int]] = [],
        ) -> tuple[Callable, list[FieldPath]]:
            """If all dependencies are field paths, then this function does nothing.
             If the dependencies contain one or more other synthetic fields, as is the
             case when chaining binary operators, then the synthetic field tree is
             flattened to a single synthetic field containing all leaf dependencies.

            Args:
              deps: Initial dependencies for synthetic field
              f: Function to apply to the values of those dependencies
              acc: Accumulator. Defaults to [].

            Returns:
              A tuple containing the potentially modified function and dependency list.
            """

            match deps:
                case []:

                    def new_f(*args):
                        new_args = []
                        _counter = 0
                        for f_, deps in acc:
                            match (f_, deps):
                                case (None, FieldPath()):
                                    new_args.append(args[_counter])
                                    _counter += 1
                                case (
                                    None,
                                    int() | float() | str() | bool() as constant,
                                ):
                                    new_args.append(constant)
                                case (f_, list() as deps):
                                    new_args.append(
                                        f_(*args[_counter : _counter + len(deps)])
                                    )
                                    _counter += len(deps)

                        return f(*new_args)

                    new_deps = []
                    for _, deps in acc:
                        match deps:
                            case FieldPath() as dep:
                                new_deps.append(dep)
                            case int() | float() | str() | bool():
                                pass
                            case list() as deps:
                                new_deps = new_deps + deps

                    return (new_f, new_deps)

                case [SyntheticField(_f=inner_f, _deps=inner_deps), *rest]:
                    acc.append((inner_f, inner_deps))
                    return mk_deps(rest, f, acc)

                case [FieldPath() as dep, *rest]:
                    acc.append((None, dep))
                    return mk_deps(rest, f, acc)

                case [int() | float() | str() | bool() as constant, *rest]:
                    acc.append((None, constant))
                    return mk_deps(rest, f, acc)

                case _ as deps:
                    raise TypeError(f"mk_deps: unexpected argument {deps}")

        (f, deps) = mk_deps(deps, f)
        self._f = f
        self._type = type_
        self._default = (
            default if default is not None else SyntheticField.default_of_type(type_)
        )
        self._deps = deps

        SyntheticField._counter += 1

    @staticmethod
    def default_of_type(type_: TypeRef.T):
        match type_.name:
            case "String":
                return ""
            case "Int":
                return 0
            case "Float":
                return 0.0
            case "Boolean":
                return False
            case _:
                return 0

    @classmethod
    def constant(cls, value: str | int | float | bool) -> SyntheticField:
        """Returns a constant ``SyntheticField`` with value ``value``.
        Useful for injecting additional static data to a schema or merging entities.

        Args:
          value: The constant field's value

        Returns:
          The constant ``SyntheticField``

        Example:

        .. code-block:: python

            >>> from subgrounds import Subgrounds, SyntheticField
            >>> sg = Subgrounds()
            >>> univ3 = sg.load_subgraph(
            ...     "https://api.thegraph.com/subgraphs/name/uniswap/uniswap-v3"
            ... )

            # Create constant SyntheticFields
            >>> univ3.Mint.tx_type = SyntheticField.constant('MINT')
            >>> univ3.Burn.tx_type = SyntheticField.constant('BURN')

            # Last 10 mints and burns
            >>> mints = univ3.Query.mints(
            ...     first=10,
            ...     orderBy=univ3.Mint.timestamp,
            ...     orderDirection='desc'
            ... )
            >>> burns = univ3.Query.burns(
            ...     first=10,
            ...     orderBy=univ3.Burn.timestamp,
            ...     orderDirection='desc'
            ... )

            # Query mints and burns. Notice that we merge the two entity tables by
            # setting `concat=True` and overwriting the column names (columns must
            # match the `FieldPaths`)
            >>> df = sg.query_df(
            ...     [
            ...         mints.transaction.id,
            ...         mints.timestamp,
            ...         mints.tx_type,
            ...         mints.origin,
            ...         mints.amountUSD,
            ...         burns.transaction.id,
            ...         burns.timestamp,
            ...         burns.tx_type,
            ...         burns.origin,
            ...         burns.amountUSD,
            ...     ],
            ...     columns=['tx_hash', 'timestamp', 'tx_type', 'origin', 'amount_USD'],
            ...     concat=True
            ..  )

            # Sort the DataFrame (output is truncated)
            >>> df.sort_values(by=['timestamp'], ascending=False)
             tx_hash   timestamp    tx_type  origin     amount_USD
            0xcbe1...  1656016553    MINT  0x3435....  7.879784e+03
            0xddda...  1656016284    MINT  0xc747....  5.110840e+04
            0xa767...  1656016284    BURN  0xd40d....  2.804573e+05
            0xc132...  1656015853    MINT  0xc747....  5.122569e+04
            0x1444...  1656015773    MINT  0xd11a....  8.897983e+03
            0x3315...  1656015693    MINT  0xb7dd....  0.000000e+00
            0xcc71...  1656015278    BURN  0xa7c4....  1.254942e+06
            0x7bbf...  1656015111    MINT  0xac56....  3.432075e+04
            0xea21...  1656014785    BURN  0x0709....  2.059106e+04
            0x3bd3...  1656014120    MINT  0x5099....  2.517578e+03
            0x1ea5...  1656014018    BURN  0x5099....  0.000000e+00
            0xb9d3...  1656013998    MINT  0x22df....  8.365750e+04
            0xc5e3...  1656013946    BURN  0xac56....  3.363809e+04
            0x7c73...  1656013913    MINT  0x4ce6....  4.837287e+04
            0x95cf...  1656013728    BURN  0x4ce6....  5.110010e+04
            0x76dd...  1656013599    MINT  0x234a....  5.363896e+03
            0x47e5...  1656013580    BURN  0xaf0f....  0.000000e+00
            0xe20e...  1656013455    BURN  0xaf0f....  0.000000e+00
            0xac3e...  1656013237    BURN  0x665d....  2.254100e+05
            0x01c3...  1656013034    BURN  0x0084....  0.000000e+00
        """

        match value:
            case str():
                return cls(lambda: value, cls.STRING, [])
            case int():
                return cls(lambda: value, cls.INT, [])
            case float():
                return cls(lambda: value, cls.FLOAT, [])
            case bool():
                return cls(lambda: value, cls.BOOL, [])

    @classmethod
    def datetime_of_timestamp(cls, timestamp: FieldPath | SyntheticField) -> Self:
        """Returns a ``SyntheticField`` that will transform the ``FieldPath``
        ``timestamp`` into a human-readable ISO8601 string.

        Args:
          timestamp: A ``FieldPath`` representing a Unix timestamp field.

        Returns:
          An ISO8601 datetime string ``SyntheticField``.

        Example:

        .. code-block:: python

            >>> from subgrounds import Subgrounds, SyntheticField
            >>> sg = Subgrounds()
            >>> univ3 = sg.load_subgraph(
            ...     "https://api.thegraph.com/subgraphs/name/uniswap/uniswap-v3"
            ... )

            # Create datetime SyntheticField
            >>> univ3.Swap.datetime = SyntheticField.datetime_of_timestamp(
            ...     univ3.Swap.timestamp
            ... )

            # Query 100 swaps
            >>> sg.query_df([
            ...     univ3.Query.swaps.timestamp,
            ...     univ3.Query.swaps.datetime,
            ... ])
                swaps_timestamp       swaps_datetime
            0        1625105710  2021-06-30 22:15:10
            1        1629253724  2021-08-17 22:28:44
            2        1647333277  2022-03-15 04:34:37
            3        1630801974  2021-09-04 20:32:54
            4        1653240241  2022-05-22 13:24:01
            ..              ...                  ...
            95       1646128326  2022-03-01 04:52:06
            96       1646128326  2022-03-01 04:52:06
            97       1626416555  2021-07-16 02:22:35
            98       1626416555  2021-07-16 02:22:35
            99       1625837291  2021-07-09 09:28:11
        """

        return SyntheticField(
            lambda timestamp: str(datetime.fromtimestamp(timestamp)),
            SyntheticField.STRING,
            timestamp,
        )

    @classmethod
    def map(
        cls,
        dict: dict[Any, Any],
        type_: TypeRef.T,
        fpath: FieldPath | SyntheticField,
        default: Any | None = None,
    ) -> Self:
        """Returns a SyntheticField that will map the values of ``fpath`` using the
        key value pairs in ``dict``. If a value is not in the dictionary, then
        ``default`` will be used instead.

        Args:
          dict: The dictionary containing the key value pairs used to map
            ``fpath``'s values
          type_: The type of the resulting field
          fpath: The FieldPath whose values will be mapped using the dictionary
          default: Default value used when a value is not in the provided dictionary

        Returns:
            A map SyntheticField

        Example:

        .. code-block:: python

            >>> from subgrounds import Subgrounds, SyntheticField
            >>> sg = Subgrounds()
            >>> univ3 = sg.load_subgraph(
            ...     'https://api.thegraph.com/subgraphs/name/uniswap/uniswap-v3'
            ... )

            # Hand-crafted mapping of pool addresses to symbol
            >>> pooladdr_symbol_map = {
            ...     '0x5777d92f208679db4b9778590fa3cab3ac9e2168': 'DAI/USDC-001',
            ...     '0x6c6bc977e13df9b0de53b251522280bb72383700': 'DAI/USDC-005',
            ...     '0x8ad599c3a0ff1de082011efddc58f1908eb6e6d8': 'USDC/ETH-030',
            ...     '0x88e6a0c2ddd26feeb64f039a2c41296fcb3f5640': 'USDC/ETH-005',
            ... }

            # Create map SyntheticField using our dictionary with 'UNKNOWN' as the
            # default value
            >>> univ3.Pool.symbol = SyntheticField.map(
            ...     pooladdr_symbol_map,
            ...     SyntheticField.STRING,
            ...     univ3.Pool.id,
            ...     'UNKNOWN'
            ... )

            # Query top 10 pools by TVL
            >>> pools = univ3.Query.pools(
            ...     first=10,
            ...     orderBy=univ3.Pool.totalValueLockedUSD,
            ...     orderDirection='desc'
            ... )
            >>> sg.query_df([
            ...     pools.id,
            ...     pools.symbol
            ... ])
                                                 pools_id  pools_symbol
            0  0xa850478adaace4c08fc61de44d8cf3b64f359bec       UNKNOWN
            1  0x5777d92f208679db4b9778590fa3cab3ac9e2168  DAI/USDC-001
            2  0x6c6bc977e13df9b0de53b251522280bb72383700  DAI/USDC-005
            3  0x8ad599c3a0ff1de082011efddc58f1908eb6e6d8  USDC/ETH-030
            4  0x88e6a0c2ddd26feeb64f039a2c41296fcb3f5640  USDC/ETH-005
            5  0x3416cf6c708da44db2624d63ea0aaef7113527c6       UNKNOWN
            6  0xcbcdf9626bc03e24f779434178a73a0b4bad62ed       UNKNOWN
            7  0xc63b0708e2f7e69cb8a1df0e1389a98c35a76d52       UNKNOWN
            8  0x4585fe77225b41b697c938b018e2ac67ac5a20c0       UNKNOWN
            9  0x4e68ccd3e89f51c3074ca5072bbac773960dfa36       UNKNOWN
        """

        return SyntheticField(
            lambda value: dict[value] if value in dict else default, type_, fpath
        )



""" Query data structure module

This module contains various data structures in the form of dataclasses that
are used to represent GraphQL queries in Subgrounds using an AST-like approach.
To the extent possible, these dataclasses are immutable (i.e.: :attr:`frozen=True`)
to enforce a functional programming style and reduce side-effects.

A typical Subgrounds request will have the following dataclass hierarchy:

.. code-block:: none

  DataRequest
  └── Document
      └── Query
          ├── VariableDefinition
          │   └── InputValue
          └── Selection
              ├── Argument
              │   └── InputValue
              └── Selection
"""

from __future__ import annotations

import logging
import warnings
from dataclasses import dataclass, field, replace
from functools import partial, reduce
from typing import Any, Callable, Iterator, Literal, Protocol, TypeVar

from pipe import map, take, traverse, where
from typing_extensions import Self

from .errors import SubgroundsError
from .schema import SchemaMeta, TypeMeta, TypeRef
from .utils import (
    extract_data,
    filter_map,
    filter_none,
    identity,
    merge,
    rel_complement,
    union,
)

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")


# ================================================================
# Query definitions, data structures and types
# ================================================================
class InputValue:
    class T(Protocol):
        @property
        def graphql(self) -> str:
            """Returns a GraphQL string representation of the input value

            Returns:
              The GraphQL string representation of the input value
            """
            ...

        @property
        def is_variable(self) -> bool:
            """Returns True i.f.f. the input value is of type Variable

            Returns:
              True i.f.f. the input value is of type Variable, otherwise False
            """
            ...

        @property
        def is_number(self) -> bool:
            """Returns True i.f.f. the input value is of type Float or Int

            Returns:
              True i.f.f. the input value is of type Float or Int, otherwise False
            """
            ...

        def iter(self) -> Iterator[InputValue.T]:
            ...

    @dataclass(frozen=True)
    class Null:
        @property
        def graphql(self) -> str:
            return "null"

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class Int:
        value: int

        @property
        def graphql(self) -> str:
            return str(self.value)

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return True

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class Float:
        value: float

        @property
        def graphql(self) -> str:
            return str(self.value)

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return True

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class String:
        value: str

        @property
        def graphql(self) -> str:
            return f'"{self.value}"'

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class Boolean:
        value: bool

        @property
        def graphql(self) -> str:
            return str(self.value).lower()

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class Enum:
        value: str

        @property
        def graphql(self) -> str:
            return self.value

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class Variable:
        name: str

        @property
        def graphql(self) -> str:
            return f"${self.name}"

        @property
        def is_variable(self) -> bool:
            return True

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self

    @dataclass(frozen=True)
    class List:
        value: list[InputValue.T]

        @property
        def graphql(self) -> str:
            return f"[{', '.join([val.graphql for val in self.value])}]"

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self
            for val in self.value:
                yield from val.iter()

    @dataclass(frozen=True)
    class Object:
        value: dict[str, InputValue.T]

        @property
        def graphql(self) -> str:
            field_strings = [
                f"{key}: {value.graphql}" for key, value in self.value.items()
            ]
            fields = ", ".join(field_strings)
            return "{" + fields + "}"

        @property
        def is_variable(self) -> bool:
            return False

        @property
        def is_number(self) -> bool:
            return False

        def iter(self) -> Iterator[InputValue.T]:
            yield self
            for val in self.value.values():
                yield from val.iter()


@dataclass(frozen=True)
class VariableDefinition:
    """Representation of a GraphQL variable definition

    Attributes:
      name: Name of the argument
      type_: GraphQL type of the argument
      default: Default value of the variable. Defaults to None.
    """

    name: str
    type_: TypeRef.T
    default: InputValue.T | None = None

    @property
    def graphql(self) -> str:
        """Returns the GraphQL string representation of the variable definition

        Example:

        >>> vardef = VariableDefinition(
        ...   name='foo',
        ...   type_=TypeRef.NonNull(TypeRef.Named(name="Int", kind="SCALAR")),
        ...   default=InputValue.Int(100)
        ... )
        >>> print(vardef.graphql)
        $foo: Int! = 100

        Returns:
          The GraphQL string representation of the variable definition
        """

        if self.default is None:
            return f"${self.name}: {TypeRef.graphql(self.type_)}"
        else:
            return (
                f"${self.name}: {TypeRef.graphql(self.type_)} = {self.default.graphql}"
            )


@dataclass(frozen=True)
class Argument:
    name: str
    value: InputValue.T

    @property
    def graphql(self) -> str:
        return f"{self.name}: {self.value.graphql}"

    def iter(self) -> Iterator[InputValue.T]:
        yield from self.value.iter()

    def iter_vars(self) -> Iterator[InputValue.Variable]:
        for iv in self.value.iter():
            match iv:
                case InputValue.Variable():
                    yield iv
                case _:
                    continue

    def for_all(self, predicate: Callable[[InputValue.T], bool]) -> bool:
        return all(self.iter() | map(predicate))

    def for_all_vars(self, predicate: Callable[[InputValue.Variable], bool]) -> bool:
        return all(self.iter_vars() | map(predicate))

    def exists(self, predicate: Callable[[InputValue.T], bool]) -> bool:
        return any(self.iter() | map(predicate))

    def exists_vars(self, predicate: Callable[[InputValue.Variable], bool]) -> bool:
        return any(self.iter_vars() | map(predicate))

    def find(self, predicate: Callable[[InputValue.T], bool]) -> InputValue.T | None:
        try:
            return next(self.iter() | where(predicate))
        except StopIteration:
            return None

    def find_var(
        self, predicate: Callable[[InputValue.Variable], bool]
    ) -> InputValue.T | None:
        try:
            return next(self.iter_vars() | where(predicate))
        except StopIteration:
            return None

    def all_defined(self, variables: Iterator[str]) -> bool:
        return self.for_all_vars(lambda var: var.name in variables)


@dataclass(frozen=True)
class Selection:
    """Represents a GraphQL field selection.

    Attributes:
      fmeta: The type definition of the field being selected.
      alias: The alias of the field selection. Defaults to None.
      arguments: The arguments, if any, of the field selection. Defaults to [].
      selection: The inner field selections, if any. Defaults to [].
    """

    fmeta: TypeMeta.FieldMeta
    alias: str | None = None
    arguments: list[Argument] = field(default_factory=list)
    selection: list[Selection] = field(default_factory=list)

    @property
    def key(self):
        if self.alias is not None:
            return self.alias
        else:
            return self.fmeta.name

    @property
    def args_graphql(self) -> str:
        if self.arguments:
            return f'({", ".join([arg.graphql for arg in self.arguments])})'
        else:
            return ""

    def graphql(self, level: int = 0) -> str:
        indent = "  " * level

        if self.alias:
            alias_str = f"{self.alias}: "
        else:
            alias_str = ""

        match (self.selection):
            case None | []:
                return f"{indent}{alias_str}{self.fmeta.name}{self.args_graphql}"
            case inner_selection:
                inner_str = "\n".join(
                    [f.graphql(level=level + 1) for f in inner_selection]
                )
                return (
                    f"{indent}{alias_str}{self.fmeta.name}{self.args_graphql}"
                    f" {{\n{inner_str}\n{indent}}}"
                )  # TODO: string templates w/ `.format`?

    @property
    def data_path(self) -> list[str]:
        match self:
            case Selection(TypeMeta.FieldMeta(name), None, _, []) | Selection(
                TypeMeta.FieldMeta(_), str() as name, _, []
            ):
                return [name]
            case Selection(
                TypeMeta.FieldMeta(name), None, _, [inner_select, *_]
            ) | Selection(TypeMeta.FieldMeta(_), str() as name, _, [inner_select, *_]):
                return [name] + inner_select.data_path

        assert False  # Suppress mypy missing return statement warning

    @property
    def data_paths(self) -> list[list[str]]:
        def f(select: Selection, keys: list[str] = []):
            match select:
                case Selection(TypeMeta.FieldMeta(name=name), None, _, []) | Selection(
                    TypeMeta.FieldMeta(name=_), str() as name, _, []
                ):
                    yield [*keys, name]
                case Selection(
                    TypeMeta.FieldMeta(name=name), None, _, inner
                ) | Selection(TypeMeta.FieldMeta(name=_), str() as name, _, inner):
                    for select in inner:
                        yield from f(select, keys=[*keys, name])

        return list(f(self))

    # ================================================================
    # Generic functions
    # ================================================================
    def iter(self) -> Iterator[Selection]:
        """Returns an iterator over all ``Selections`` of the current selection tree."""

        yield self
        for inner in self.selection:
            yield from inner.iter()

    def iter_args(self, recurse: bool = True) -> Iterator[Argument]:
        """Returns an iterator over all ``Arguments`` of the current ``Selection``.

        If ``recurse == True``, then the iterator also includes ``Arguments`` of
        inner ``Selections``.
        """

        for arg in self.arguments:
            yield arg

        if recurse:
            for inner in self.selection:
                yield from inner.iter_args()

    def filter(self, predicate: Callable[[Selection], bool]) -> Selection | None:
        """Returns a new ``Selection`` object containing all attributes of the current
        ``Selection`` if ``predicate(self) == True`` and ``None`` otherwise. The
        function if also applied recursively to inner ``Selections``.

        Args:
          predicate: _description_

        Returns:
          _description_
        """

        if predicate(self):
            return Selection(
                fmeta=self.fmeta,
                alias=self.alias,
                arguments=self.arguments,
                selection=list(
                    self.selection
                    | filter_map(partial(Selection.filter, predicate=predicate))
                ),
            )

        return None

    def filter_args(
        self, predicate: Callable[[Argument], bool], recurse: bool = True
    ) -> Selection:
        """Returns a new ``Selection`` object which contains all attributes of
        the current ``Selection`` except for ``Arguments`` for which
        ``predicate(arg) == True``.

        If ``recurse == True``, then the function is applied recursively to
        inner ``Selections``

        Args:
          predicate: _description_
          recurse: _description_. Defaults to True.

        Returns:
          _description_
        """
        return Selection(
            fmeta=self.fmeta,
            alias=self.alias,
            arguments=list(self.arguments | where(predicate)),
            selection=list(
                self.selection
                | map(partial(Selection.filter_args, predicate=predicate))
            )
            if recurse
            else self.selection,
        )

    def map(
        self,
        map_f: Callable[[Selection], Selection],
        priority: Literal["self"] | Literal["children"] = "self",
    ) -> Selection:
        """Returns a new ``Selection`` object containing the same selection tree
        as the current ``Selection`` where each ``Selection`` object ``s`` is
        ``map_f(s)``

        Args:
          map_f: Mapping function to apply to each ``Selection``

        Returns:
          _description_
        """

        match priority:
            case "self":
                new_selection = map_f(self)

                return Selection(
                    fmeta=new_selection.fmeta,
                    alias=new_selection.alias,
                    arguments=new_selection.arguments,
                    selection=list(
                        self.selection
                        | map(partial(Selection.map, map_f=map_f, priority=priority))
                    ),
                )

            case "children":
                new_selection = Selection(
                    fmeta=self.fmeta,
                    alias=self.alias,
                    arguments=self.arguments,
                    selection=list(
                        self.selection
                        | map(partial(Selection.map, map_f=map_f, priority=priority))
                    ),
                )

                return map_f(new_selection)

            case _:
                raise SubgroundsError(f"map: invalid priority {priority}")

    def map_args(
        self,
        map_f: Callable[[Argument], Argument | list[Argument]],
        recurse: bool = True,
    ) -> Selection:
        """Replaces each ``Argument`` ``arg`` in the current ``Selection`` with
         ``map_f(arg)`` and returns a new ``Selection`` object containing
         the modified arguments.

        If ``recurse == True``, then the function is applied recursively to inner
        ``Selections``.

        Args:
          map_f: _description_
          recurse: _description_. Defaults to True.

        Returns:
          _description_
        """

        return Selection(
            fmeta=self.fmeta,
            alias=self.alias,
            arguments=list(self.arguments | map(map_f)),
            selection=list(
                self.selection
                | map(partial(Selection.map_args, map_f=map_f))
                | traverse
            )
            if recurse
            else self.selection,
        )

    def filter_map(
        self, map_f: Callable[[Selection], Selection | None]
    ) -> Selection | None:
        new_selection = map_f(self)

        if new_selection is not None:
            return Selection(
                fmeta=new_selection.fmeta,
                alias=new_selection.alias,
                arguments=new_selection.arguments,
                selection=list(
                    self.selection
                    | filter_map(partial(Selection.filter_map, map_f=map_f))
                ),
            )

        return None

    def filter_map_args(
        self,
        map_f: Callable[[Argument], Argument | list[Argument | None]],
        recurse: bool = True,
    ) -> Selection:
        return Selection(
            fmeta=self.fmeta,
            alias=self.alias,
            arguments=list(self.arguments | filter_map(map_f) | traverse),
            selection=list(
                self.selection | map(partial(Selection.filter_map_args, map_f=map_f))
            )
            if recurse
            else self.selection,
        )

    def for_all(self, predicate: Callable[[Selection], bool]) -> bool:
        if predicate(self):
            return all(
                self.selection | map(partial(Selection.for_all, predicate=predicate))
            )

        return False

    def for_all_args(
        self, predicate: Callable[[Argument], bool], recurse: bool = True
    ) -> bool:
        if all(self.arguments | map(predicate)):
            if recurse:
                return all(
                    self.selection
                    | map(partial(Selection.for_all_args, predicate=predicate))
                )

            return True

        return False

    def exists(self, predicate: Callable[[Selection], bool]) -> bool:
        if predicate(self):
            return True

        return any(self.selection | map(partial(Selection.exists, predicate=predicate)))

    def exists_args(
        self, predicate: Callable[[Argument], bool], recurse: bool = True
    ) -> bool:
        if any(self.arguments | map(predicate)):
            return True

        if recurse:
            return any(
                self.selection
                | map(partial(Selection.exists_args, predicate=predicate))
            )

        return False

    def find(self, predicate: Callable[[Selection], bool]) -> Selection | None:
        try:
            return next(self.iter() | where(predicate))
        except StopIteration:
            return None

    def find_args(
        self, predicate: Callable[[Argument], bool], recurse: bool = True
    ) -> Argument | None:
        try:
            return next(self.iter_args(recurse=recurse) | where(predicate))
        except StopIteration:
            return None

    def find_all(self, predicate: Callable[[Selection], bool]) -> list[Selection]:
        return list(self.iter() | where(predicate))

    def find_all_args(
        self, predicate: Callable[[Argument], bool], recurse: bool = True
    ) -> list[Argument]:
        return list(self.iter_args(recurse=recurse) | where(predicate))

    T = TypeVar("T")

    def fold(
        self,
        fold_f: Callable[[Selection, list[Selection], list[T]], T | list[T]],
        parents: list[Selection] = [],
    ) -> T | list[T]:
        inner = list(
            self.selection
            | map(partial(Selection.fold, fold_f=fold_f, parents=[*parents, self]))
            | traverse
        )
        return fold_f(self, parents, inner)

    def contains_list(self) -> bool:
        """Returns True i.f.f. the selection :attr:`self` selects a field of type
        list.

        Args:
          self: The selection to traverse

        Returns:
          True if selection or nested selections selects a list field. False otherwise.
        """

        return self.exists(lambda select: select.fmeta.type_.is_list)

    def split(self) -> list[Selection]:
        """Returns a list of selections where each of the selections corresponds
        to a single selection path from the root to a leaf for each leaf selected
        in :attr:`self`.

        Example (simplified, does not show all attributes):

        >>> select = Selection('foo', inner=[
        ...   Selection('bar', inner=[
        ...     Selection('field0', inner=[]),
        ...     Selection('field1', inner=[]),
        ...   ]),
        ...   Selection('x', inner=[])
        ... ])
        >>> split(select)
        [
          Selection(
            'foo', inner=[Selection('bar', inner=[Selection('field0', inner=[])])]),
          Selection(
            'foo', inner=[Selection('bar', inner=[Selection('field1', inner=[])])]),
          Selection(
            'foo', inner=[Selection('x', inner=[])]),
        ]

        Args:
          self: The selection to split

        Returns:
          The split selections
        """

        match self:
            case Selection(_, _, _, [] | None):
                return [self]
            case Selection(fmeta, alias, args, inner_select):
                return list(
                    inner_select
                    | map(Selection.split)
                    | traverse
                    | map(
                        lambda inner_select: Selection(fmeta, alias, args, inner_select)
                    )
                )

    def extract_data(self: Selection, data: dict | list[dict]) -> list[Any] | Any:
        return extract_data(self.data_path, data)

    def add(self, new_selections: Selection | list[Selection]) -> Selection:
        """Returns a new selection consisting of a copy of :attr:`self` expanded
        with the selection(s) :attr:`new_selections`. It is assumed that
        :attr:`new_selections` are inner selections of the root selection
        :attr:`self`.

        Args:
          self: The Selection object to be expanded
          new_selections: A single or multiple Selection object(s) to be added to
            :attr:`self`

        Returns:
          The resulting new selection, i.e.: :attr:`self` expanded with
            :attr:`new_selections`
        """

        match new_selections:
            case Selection() as new_selection:
                return self.add([new_selection])

            case list() as new_selections:
                return Selection(
                    fmeta=self.fmeta,
                    alias=self.alias,
                    arguments=self.arguments,
                    selection=union(
                        self.selection,
                        new_selections,
                        key=lambda select: select.fmeta.name,
                        combine=Selection.combine,
                    ),
                )

    def remove(self, to_remove: Selection | list[Selection]) -> Selection:
        """Returns a new Selection object consisting of a copy of :attr:`self`
        without the selections in :attr:`selections_to_remove`.

        Args:
          to_remove: The selection(s) to remove from :attr:`self`

        Returns:
          The new trimmed down selection, i.e.: :attr:`self` without
            :attr:`selections_to_remove`
        """

        def combine(
            select: Selection, selection_to_remove: Selection
        ) -> Selection | None:
            if selection_to_remove.selection == []:
                return None
            else:
                return select.remove(selection_to_remove.selection)

        # TODO: Resolve mypy complaints
        match to_remove:
            case Selection() as to_remove:
                return self.remove([to_remove])

            case list() as to_remove:
                return Selection(
                    fmeta=self.fmeta,
                    alias=self.alias,
                    arguments=self.arguments,
                    selection=filter_none(
                        union(
                            self.selection,
                            to_remove,
                            key=lambda s: s.fmeta.name,
                            combine=combine,
                        )
                    ),
                )

    def variable_args(self, recurse: bool = True) -> Iterator[Argument]:
        """Returns all arguments in the current selection which have been given a
        variable as value.

        If ``recurse == True``, then the function is applied recursively to inner
        selections.

        Args:
          recurse: _description_. Defaults to True.

        Returns:
          _description_
        """
        for arg in self.iter_args(recurse=recurse):
            if type(arg.value) == InputValue.Variable:
                yield arg

    def infer_variable_definitions(self: Selection) -> list[VariableDefinition]:
        def vardef_of_arg(arg):
            match arg:
                case Argument(name=argname, value=InputValue.Variable(name=varname)):
                    return VariableDefinition(varname, self.fmeta.type_of(argname))

                case Argument(name=argname, value=InputValue.Object(fields)):
                    return list(
                        fields.values()
                        | where(lambda iv: type(iv) == InputValue.Variable)
                        | map(
                            lambda iv: VariableDefinition(
                                iv.name, self.fmeta.type_of(argname)
                            )
                        )
                    )

                case _:
                    return []

        var_defs = list(self.arguments | map(vardef_of_arg) | traverse)

        return (
            list(self.selection | map(Selection.infer_variable_definitions) | traverse)
            + var_defs
        )

    def combine(self: Selection, other: Selection) -> Selection:
        if self.key != other.key:
            raise SubgroundsError(f"Selection.combine: {self.key} != {other.key}")

        return Selection(
            fmeta=self.fmeta,
            alias=self.alias,
            arguments=self.arguments,
            selection=filter_none(
                union(
                    self.selection,
                    other.selection,
                    key=lambda select: select.fmeta.name,
                    combine=Selection.combine,
                )
            ),
        )

    @staticmethod
    def merge(selections: list[Selection]) -> list[Selection]:
        """Returns a list of Selection objects resulting from merging
        :attr:`selections` to the extent possible.

        Args:
          selections The selections to be merged

        Returns:
          _description_
        """

        def f(selections: list[Selection], other: Selection) -> list[Selection]:
            try:
                next(selections | where(lambda select: select.key == other.key))
                return list(
                    selections
                    | map(
                        lambda select: Selection.combine(select, other)
                        if select.key == other.key
                        else select
                    )
                )
            except StopIteration:
                return selections + [other]

        return reduce(f, selections, [])

    def contains(self, other: Selection) -> bool:
        """Returns True i.f.f. the Selection :attr:`other` is a subtree of the
        Selection :attr:`self` and False otherwise

        Args:
          self: The selection
          other: The subselection

        Returns:
          True i.f.f. :attr:`other` is in :attr:`self`
        """

        if (
            self.fmeta == other.fmeta
            and rel_complement(
                other.selection, self.selection, key=lambda s: s.fmeta.name
            )
            == []
        ):
            return all(
                other.selection
                | map(
                    lambda s: Selection.contains(
                        next(
                            filter(
                                lambda s_: s.fmeta.name == s_.fmeta.name, self.selection
                            )
                        ),
                        s,
                    )
                )
            )
        else:
            return False

    def contains_argument(self, argname: str, recurse: bool = True) -> bool:
        """Returns True i.f.f. there is an Argument object in :attr:`self` named
        :attr:`argname`. If :attr:`recurse` is True, then the method also checks the
        nested selections for an argument named :attr:`argname`.

        Args:
          argname: The name of the argument
          recurse: Flag indicating whether or not the method should be run recursively
            on nested selections. Defaults to True.

        Returns:
          True i.f.f. there is an argument named :attr:`argname` in :attr:`self`
        """

        return self.exists_args(lambda arg: arg.name == argname, recurse=recurse)

    def get_argument(self, argname: str, recurse: bool = True) -> Argument | None:
        """Returns an Argument object corresponding to the argument in the Selection
        object :attr:`select` with name :attr:`argname`. If :attr:`select` does not
        contain such an argument and :attr:`recurse` is True, then the function is
        called recursively on :attr:`select`'s inner selections. If no such argument
        is found in :attr:`select` or its inner selections, then the function raises
        an exception.

        Args:
          argname: The name of the argument to find
          recurse: Flag indicating whether or not the method should be run recursively
            on nested selections. Defaults to True.

        Raises:
          KeyError: If no argument named :attr:`argname` exists in the selection
            :attr:`self`.

        Returns:
          The argument in :attr:`select` with name :attr:`argname` (if any).
        """

        return self.find_args(lambda arg: arg.name == argname, recurse=recurse)

    def get_argument_by_variable(
        self, varname: str, recurse: bool = True
    ) -> Argument | None:
        """Returns an Argument object corresponding to the argument in the Selection
        object :attr:`select` whose value is a variable named :attr:`varname`. If
        :attr:`select` does not contain such an argument and :attr:`recurse` is True,
        then the function is called recursively on :attr:`select`'s inner selections.
        If no such argument is found in :attr:`select` or its inner selections, then
        the function raises an exception

        Args:
          varname: The name of the variable to find
          recurse: Flag indicating whether or not the method should be run recursively.
            Defaults to True.

        Raises:
          KeyError: If no argument with variable value named :attr:`varname` exists
            in the selection :attr:`self`.

        Returns:
          The argument in :attr:`select` with variable value named :attr:`varname`
            if it exists
        """

        return self.find_args(
            lambda arg: arg.exists_vars(lambda var: var.name == varname),
            recurse=recurse,
        )

    # TODO: Replace substitute_arg calls by map_args call
    def substitute_arg(
        self, argname: str, replacement: Argument | list[Argument], recurse: bool = True
    ) -> Selection:
        """Returns a new Selection object containing the same data as :attr:`self`
        with the argument named :attr:`argname` replaced with :attr:`replacement`.
        If :attr:`recurse` is True, then the method is called recursively on
        :attr:`self`'s inner selections and the substitution is also applied to the
        latter.

        Args:
          argname: The name of the argument to substitute.
          replacement: The argument(s) replacement
          recurse: Flag indicating whether or not the method should be run recursively.
            Defaults to True.

        Returns:
          _description_
        """

        return self.map_args(
            lambda arg: replacement if arg.name == argname else arg, recurse=recurse
        )

    def select(self: Selection, other: Selection) -> Selection:
        if other.selection == []:
            return self
        else:
            return Selection(
                fmeta=self.fmeta,
                alias=self.alias,
                arguments=self.arguments,
                selection=list(
                    other.selection
                    | map(
                        lambda s: next(
                            self.selection
                            | where(lambda s_: s_.fmeta.name == s.fmeta.name)
                            | map(lambda s_: Selection.select(s_, s))
                            | take(1)
                        )
                    )
                ),
            )

    def prune_undefined(self, variables: Iterator[str]) -> Selection | None:
        """Return a new ``Selection`` containing the subtree of the current
        ``Selection`` where all argument ``InputValues`` are defined,
        i.e.: each argument's ``InputValue`` is either
            1) not of type ``InputValue.Variable``
            or
            2) of type ``InputValue.Variable``
            and
            the variable name is contained in ``variables``.

        Args:
          variables: An iterator over defined variables

        Returns:
          A new pruned ``Selection`` object
        """

        return self.filter(
            lambda select: select.for_all_args(
                lambda arg: arg.all_defined(variables), recurse=False
            )
        )


@dataclass(frozen=True)
class Query:
    name: str | None = None
    selection: list[Selection] = field(default_factory=list)

    # Variables as query arguments, not the values of those variables
    variables: list[VariableDefinition] = field(default_factory=list)

    @property
    def graphql(self) -> str:
        """Returns a string containing a GraphQL query matching the current query

        Returns:
          The string containing the GraphQL query
        """

        selection_str = "\n".join(
            [select.graphql(level=1) for select in self.selection]
        )

        if len(self.variables) > 0:
            args_str = f'({", ".join([vardef.graphql for vardef in self.variables])})'
        else:
            args_str = ""

        return f"query{args_str} {{\n{selection_str}\n}}"

    # ================================================================
    # Generic functions
    # ================================================================
    def iter(self) -> Iterator[Selection]:
        """Returns an iterator over all ``Selections`` of the selection tree of
        the current ``Query``."""

        for select in self.selection:
            yield from select.iter()

    def iter_args(self) -> Iterator[Argument]:
        """Returns an iterator over all ``Arguments`` of the selection tree of
        the current ``Query``."""

        for select in self.selection:
            yield from select.iter_args()

    def iter_vardefs(self) -> Iterator[VariableDefinition]:
        """Returns an iterator over all ``VariableDefinitions`` of the selection tree of
        the current ``Query``."""

        for vardef in self.variables:
            yield vardef

    def filter(self, predicate: Callable[[Selection], bool]) -> Query:
        """Returns a new ``Query`` object containing all selections ``s`` that satisfy
        ``predicate(s) == True``.

        Args:
          predicate: _description_

        Returns:
          _description_
        """

        return Query(
            name=self.name,
            selection=list(
                self.selection
                | filter_map(partial(Selection.filter, predicate=predicate))
            ),
            variables=self.variables,
        )

    def filter_args(self, predicate: Callable[[Argument], bool]) -> Query:
        """Returns a new ``Query`` object containing all selections arguments ``arg``
        that satisfy ``predicate(arg) == True``.

        Args:
          predicate: _description_

        Returns:
          _description_
        """

        return Query(
            name=self.name,
            selection=list(
                self.selection
                | map(partial(Selection.filter_args, predicate=predicate))
            ),
            variables=self.variables,
        )

    def filter_vardefs(self, predicate: Callable[[VariableDefinition], bool]) -> Query:
        return Query(
            name=self.name,
            selection=self.selection,
            variables=list(self.variables | where(predicate)),
        )

    def map(
        self,
        map_f: Callable[[Selection], Selection],
        priority: Literal["self"] | Literal["children"] = "self",
    ) -> Query:
        """Applies the function ``map_f`` to each ``Selection`` in the current
         ``Query`` and returns a new ``Query`` object containing the resulting
         ``Selections``.

        Args:
          map_f: Mapping function to apply to each ``Selection``

        Returns:
          _description_
        """

        return Query(
            name=self.name,
            selection=list(
                self.selection
                | map(partial(Selection.map, map_f=map_f, priority=priority))
            ),
            variables=self.variables,
        )

    def map_args(self, map_f: Callable[[Argument], Argument]) -> Query:
        """Applies the function ``map_f`` to each ``Argument`` in the current
         ``Query`` and returns a new ``Query`` object containing the resulting
         ``Arguments``.

        Args:
          map_f: _description_

        Returns:
          _description_
        """

        return Query(
            name=self.name,
            selection=list(
                self.selection | map(partial(Selection.map_args, map_f=map_f))
            ),
            variables=self.variables,
        )

    def map_vardefs(
        self, map_f: Callable[[VariableDefinition], VariableDefinition]
    ) -> Query:
        return Query(
            name=self.name,
            selection=self.selection,
            variables=list(self.variables | map(map_f)),
        )

    def filter_map(self, map_f: Callable[[Selection], Selection | None]) -> Query:
        return Query(
            name=self.name,
            selection=list(
                self.selection | filter_map(partial(Selection.filter_map, map_f=map_f))
            ),
            variables=self.variables,
        )

    def filter_map_args(self, map_f: Callable[[Argument], Argument | None]) -> Query:
        return Query(
            name=self.name,
            selection=list(
                self.selection | map(partial(Selection.filter_map_args, map_f=map_f))
            ),
            variables=self.variables,
        )

    def filter_map_vardefs(
        self, map_f: Callable[[VariableDefinition], VariableDefinition | None]
    ) -> Query:
        return Query(
            name=self.name,
            selection=self.selection,
            variables=list(self.variables | filter_map(map_f)),
        )

    def for_all(self, predicate: Callable[[Selection], bool]) -> bool:
        return all(
            self.selection | map(partial(Selection.for_all, predicate=predicate))
        )

    def for_all_args(self, predicate: Callable[[Argument], bool]) -> bool:
        return all(
            self.selection | map(partial(Selection.for_all_args, predicate=predicate))
        )

    def for_all_vardefs(self, predicate: Callable[[VariableDefinition], bool]) -> bool:
        return all(self.variables | map(predicate))

    def exists(self, predicate: Callable[[Selection], bool]) -> bool:
        return any(self.selection | map(partial(Selection.exists, predicate=predicate)))

    def exists_args(self, predicate: Callable[[Argument], bool]) -> bool:
        return any(
            self.selection | map(partial(Selection.exists_args, predicate=predicate))
        )

    def exists_vardefs(self, predicate: Callable[[VariableDefinition], bool]) -> bool:
        return any(self.variables | map(predicate))

    def find(self, predicate: Callable[[Selection], bool]) -> Selection | None:
        try:
            return next(self.iter() | where(predicate))
        except StopIteration:
            return None

    def find_args(self, predicate: Callable[[Argument], bool]) -> Argument | None:
        try:
            return next(self.iter_args() | where(predicate))
        except StopIteration:
            return None

    def find_vardefs(
        self, predicate: Callable[[VariableDefinition], bool]
    ) -> VariableDefinition | None:
        try:
            return next(self.iter_vardefs() | where(predicate))
        except StopIteration:
            return None

    T = TypeVar("T")

    def fold(
        self,
        fold_f: Callable[[Selection, list[Selection], list[T]], T | list[T]],
    ) -> list[T]:
        return list(self.selection | map(partial(Selection.fold, fold_f=fold_f)))

    def infer_variable_definitions(self: Query) -> list[VariableDefinition]:
        return list(
            self.selection | map(Selection.infer_variable_definitions) | traverse
        )

    def add(self: Query, other: Query | Selection | list[Selection]) -> Query:
        """Returns a new Query containing all selections in :attr:'self' along with
        the new selections in :attr:`other`

        Args:
          self: The query to which new selection(s) or query are to be added
          other: The new selection(s) or query to be added to the query

        Returns:
          A new `Query` objects containing all selections
        """

        match other:
            case Selection() as new_selection:
                return self.add([new_selection])

            case Query() as query:
                return self.add(query.selection)

            case list() as new_selections:
                return Query(
                    name=self.name,
                    selection=union(
                        self.selection,
                        new_selections,
                        key=lambda select: select.key,
                        combine=Selection.combine,
                    ),
                )

    def add_vardefs(self, vardefs: list[VariableDefinition]) -> Query:
        return Query(
            name=self.name,
            selection=self.selection,
            variables=union(
                self.variables,
                vardefs,
                key=lambda vardef: vardef.name,
                combine=lambda _, x: x,
            ),
        )

    def remove(self: Query, other: Query | Selection | list[Selection]) -> Query:
        """Returns a new :class:`Query` object containing all selections in
        :attr:`self` minus the subquery or selection(s) specified in :attr:`other`.

        Note: :attr:`other` does not need to be a "full" selection (i.e.: a
        selection all the way to leaves of the GraphQL schema).

        Example:

        >>> og_selection = Selection(
        ...     TypeMeta.FieldMeta(
        ...         'pair',
        ...         description="",
        ...         args=[],
        ...         type=TypeRef.non_null_list("Pair", kind="OBJECT")
        ...     ),
        ...     None,
        ...     [],
        ...     [
        ...         Selection(
        ...             TypeMeta.FieldMeta(
        ...                 'token0',
        ...                 description="",
        ...                 args=[],
        ...                 type=TypeRef.Named(name="Token", kind="OBJECT")
        ...             ),
        ...             None,
        ...             [],
        ...             [
        ...                 Selection(
        ...                     TypeMeta.FieldMeta(
        ...                         'id',
        ...                         description="",
        ...                         args=[],
        ...                         type=TypeRef.Named(name="String", kind="SCALAR")
        ...                     ),
        ...                     None,
        ...                     [],
        ...                     []
        ...                 ),
        ...                 Selection(
        ...                     TypeMeta.FieldMeta(
        ...                         'name',
        ...                         description="",
        ...                         args=[],
        ...                         type=TypeRef.Named(name="String", kind="SCALAR")
        ...                     ),
        ...                     None,
        ...                     [],
        ...                     []
        ...                 ),
        ...                 Selection(
        ...                     TypeMeta.FieldMeta(
        ...                         'symbol',
        ...                         description="",
        ...                         args=[],
        ...                         type=TypeRef.Named(name="String", kind="SCALAR")
        ...                     ),
        ...                     None,
        ...                     [],
        ...                     []
        ...                 ),
        ...             ]
        ...         )
        ...     ]
        ... )
        >>> selection_to_remove = Selection(
        ...     TypeMeta.FieldMeta(
        ...         'token0',
        ...         description="",
        ...         args=[],
        ...         type=TypeRef.Named(name="Token", kind="OBJECT")
        ...     ),
        ...     None,
        ...     [],
        ...     []
        ... )
        >>> og_selection.remove(selection_to_remove)
        Selection(
            TypeMeta.FieldMeta(
                'pair',
                description="",
                args=[],
                type=TypeRef.non_null_list("Pair", kind="OBJECT")
            ),
            None,
            [],
            []
        )

        Args:
          query (Query): The query to which a selection has to be removed
          other (Query | Selection | list[Selection]): The subquery or
            selection(s) to remove from :attr:`self`

        Returns:
          Query: A new `Query` object containing the original query selections
            minus :attr:`other`
        """

        def combine(
            select: Selection, selection_to_remove: Selection
        ) -> Selection | None:
            if selection_to_remove.selection == []:
                return None
            else:
                return select.remove(selection_to_remove.selection)

        match other:
            case Selection() as to_remove:
                return self.remove([to_remove])

            case Query() as to_remove:
                return self.remove(to_remove.selection)

            case list() as to_remove:
                return Query(
                    name=self.name,
                    selection=filter_none(
                        union(
                            self.selection,
                            to_remove,
                            key=lambda s: s.fmeta.name,
                            combine=combine,
                        )
                    ),
                    variables=self.variables,
                )

    @staticmethod
    def transform(
        query: Query,
        variable_f: Callable[[VariableDefinition], VariableDefinition] = identity,
        selection_f: Callable[[Selection], Selection] = identity,
    ) -> Query:
        return reduce(
            Query.add,
            query.selection | map(selection_f) | traverse,
            Query(
                name=query.name,
                variables=list(query.variables | map(variable_f) | traverse),
            ),
        )

    def contains_selection(self: Query, selection: Selection) -> bool:
        """Returns True i.f.f. the selection tree :attr:`selection` is present in
        :attr:`query`.

        Args:
          query (Query): A query object
          selection (Selection): The selection to be found (or not) in :attr:`query`

        Returns:
          bool: True if the :attr:`selection` is present in :attr:`query`, False
            otherwise.
        """

        return any(
            self.selection | map(lambda select: Selection.contains(select, selection))
        )

    def contains_argument(self, argname: str) -> bool:
        return self.exists_args(lambda arg: arg.name == argname)

    def get_argument(self, argname: str) -> Argument | None:
        return self.find_args(lambda arg: arg.name == argname)

    # TODO: Replace substitute_arg calls by map_args call
    @staticmethod
    def substitute_arg(
        query: Query, arg_name: str, replacement: Argument | list[Argument]
    ) -> Query:
        return Query(
            name=query.name,
            selection=list(
                query.selection
                | map(
                    partial(
                        Selection.substitute_arg,
                        arg_name=arg_name,
                        replacement=replacement,
                    )
                )
            ),
            variables=query.variables,
        )

    # TODO: Cleanup select
    @staticmethod
    def contains(query: Query, other: Query) -> bool:
        """Returns True i.f.f. all selections in `other` are contained in `query`.
        In other words, returns true i.f.f. `other` is a subset of `query`.

        Note: `other` does not need to include "full" selections
        (i.e.: selections all the way to leaves of the GraphQL schema).

        Args:
          query (Query): The query that is to be checked
          other (Query): The query that has to be in `query`

        Returns:
          bool: True i.f.f. all selections in `other` are contained in `query`,
           otherwise False
        """

        return all(other.selection | map(partial(Query.contains_selection, query)))

    # TODO: Cleanup select
    @staticmethod
    def select(query: Query, other: Query) -> Query:
        """Returns a new Query

        Args:
            query (Query): [description]
            other (Query): [description]

        Returns:
            Query: [description]
        """

        return Query(
            name=query.name,
            selection=list(
                other.selection
                | map(
                    lambda s: next(
                        query.selection
                        | where(lambda s_: s_.fmeta.name == s.fmeta.name)
                        | map(lambda s_: Selection.select(s_, s))
                        | take(1)
                    )
                )
            ),
            variables=query.variables,
        )

    def prune_undefined(self, variables: Iterator[str]) -> Query:
        return self.filter_map(
            partial(Selection.prune_undefined, variables=variables)
        ).filter_vardefs(lambda vardef: vardef.name in variables)


@dataclass(frozen=True)
class Fragment:
    name: str
    type_: TypeRef.T
    selection: list[Selection] = field(default_factory=list)

    # Variables as fragment arguments, not the values of those variables
    variables: list[VariableDefinition] = field(default_factory=list)

    @property
    def graphql(self):
        selection_str = "\n".join(select.graphql(level=1) for select in self.selection)
        root_type_name = TypeRef.root_type_name(self.type_)
        return f"fragment {self.name} on {root_type_name} {{\n{selection_str}\n}}"

    # TODO: Cleanup combine
    @staticmethod
    def combine(frag: Fragment, other: Fragment) -> Fragment:
        raise NotImplementedError("Fragment.combine")

    # TODO: Cleanup transform
    @staticmethod
    def transform(frag: Fragment, f: Callable[[Selection], Selection]) -> Fragment:
        return Fragment(
            name=frag.name, type_=frag.type_, selection=list(frag.selection | map(f))
        )


@dataclass(frozen=True)
class Document:
    url: str
    query: Query
    fragments: list[Fragment] = field(default_factory=list)

    # A list of variable assignments. For non-repeating queries
    # the list would be of length 1 (i.e.: only one set of query variable
    # assignments)
    variables: dict[str, Any] = field(default_factory=dict)

    @property
    def graphql(self):
        return "\n".join(
            [self.query.graphql, *list(self.fragments | map(lambda frag: frag.graphql))]
        )

    @staticmethod
    def mk_single_query(url: str, query: Query) -> Document:
        return Document(url, query)

    def filter(self, predicate: Callable[[Selection], bool]) -> Document:
        return Document(
            url=self.url,
            query=self.query.filter(predicate),
            fragments=self.fragments,  # TODO: Add filtering to fragments
            variables=self.variables,
        )

    def filter_args(self, predicate: Callable[[Argument], bool]) -> Document:
        return Document(
            url=self.url,
            query=self.query.filter_args(predicate),
            fragments=self.fragments,  # TODO: Add filtering to fragments
            variables=self.variables,
        )

    def map(self, map_f: Callable[[Selection], Selection]) -> Document:
        """Applies the function ``map_f`` to each ``Selection`` in the current
        ``Document`` and returns a new ``Document`` object containing the
        resulting ``Selections``.

        Args:
            map_f (Callable[[Selection], Selection]): Mapping function to apply
              to each ``Selection``

        Returns:
            Query: _description_
        """

        return Document(
            url=self.url,
            query=self.query.map(map_f),
            fragments=self.fragments,  # TODO: Add mapping to fragments
            variables=self.variables,
        )

    def map_args(self, map_f: Callable[[Argument], Argument]) -> Document:
        """Applies the function ``map_f`` to each ``Argument`` in the current
        ``Document`` and returns a new ``Document`` object containing the resulting
        ``Arguments``.

        Args:
            map_f (Callable[[Argument], Argument]): _description_

        Returns:
            Selection: _description_
        """
        return Document(
            url=self.url,
            query=self.query.map_args(map_f),
            fragments=self.fragments,  # TODO: Add mapping to fragments
            variables=self.variables,
        )

    def filter_map(self, map_f: Callable[[Selection], Selection | None]) -> Document:
        return Document(
            url=self.url,
            query=self.query.filter_map(map_f),
            fragments=self.fragments,  # TODO: Add mapping to fragments
            variables=self.variables,
        )

    # TODO: Cleanup combine
    @staticmethod
    def combine(doc: Document, other: Document) -> Document:
        return Document(
            url=doc.url,
            query=doc.query.combine(other.query),
            fragments=union(
                doc.fragments,
                other.fragments,
                key=lambda frag: frag.name,
                combine=Fragment.combine,
            ),
        )

    # TODO: Cleanup transform
    @staticmethod
    def transform(
        doc: Document,
        query_f: Callable[[Query], Query] = identity,
        fragment_f: Callable[[Fragment], Fragment] = identity,
    ) -> Document:
        return Document(
            url=doc.url,
            query=query_f(doc.query),
            fragments=list(doc.fragments | map(fragment_f)),
            variables=doc.variables,
        )

    def prune_undefined(self, variables: Iterator[str]) -> Document:
        """Returns a new ``Document`` object that contains the subset of the current
        ``Document``'s query containing only the ``Selections`` for which all its
        arguments are defined (i.e.: either constants or variables in ``variables``).

        Args:
            variables (dict[str, Any]): _description_

        Returns:
            Document: _description_
        """
        return Document(
            url=self.url,
            query=self.query.prune_undefined(variables),
            fragments=self.fragments,
            variables=variables,
        )


@dataclass(frozen=True)
class DocumentResponse:
    # TODO: update to recursive json dict w/ pydantic 2.0
    url: str
    data: dict[str, Any] = field(default_factory=dict)

    def combine(self: Self, other: Self):
        return replace(self, data=merge(self.data, other.data))


@dataclass(frozen=True)
class DataRequest:
    documents: list[Document] = field(default_factory=list)

    @property
    def graphql(self):
        return "\n".join(list(self.documents | map(lambda doc: doc.graphql)))

    @classmethod
    def combine(cls, req: DataRequest, other: DataRequest):
        return cls(
            documents=union(
                req.documents,
                other.documents,
                key=lambda doc: doc.url,
                combine=Document.combine,
            )
        )

    @classmethod
    def transform(cls, req: DataRequest, f: Callable[[Document], Document]):
        return cls(documents=list(req.documents | map(f)))

    @classmethod
    def single_query(cls, url: str, query: Query):
        return cls([Document(url, query)])

    @classmethod
    def single_document(cls, doc: Document):
        return cls([doc])

    def add_documents(self, docs: Document | list[Document]):
        return replace(self, documents=list((self.documents, docs) | traverse))


@dataclass(frozen=True)
class DataResponse:
    responses: list[DocumentResponse] = field(default_factory=list)

    def add_responses(self, resps: DocumentResponse | list[DocumentResponse]):
        return replace(self, responses=list((self.responses, resps) | traverse))


# ================================================================
# Utility functions
# ================================================================
def selections_of_object(
    schema: SchemaMeta, object_: TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta
):
    """Returns generator of Selection objects that selects all non-list fields of
    GraphQL Object of Interface :attr:`object_`.

    Args:
      schema (SchemaMeta): _description_
      object_ (TypeMeta.ObjectMeta | TypeMeta.InterfaceMeta): _description_

    Yields:
      _type_: _description_
    """
    for fmeta in object_.fields:
        if not fmeta.type_.is_list and len(fmeta.arguments) == 0:
            match schema.type_of_typeref(fmeta.type_):
                case TypeMeta.ObjectMeta() | TypeMeta.InterfaceMeta() as inner_object:
                    yield Selection(
                        fmeta, selection=[Selection(inner_object.field("id"))]
                    )
                case _:
                    yield Selection(fmeta)


def input_value_of_argument(
    schema: SchemaMeta, argmeta: TypeMeta.ArgumentMeta, value: Any
) -> InputValue:
    def fmt_value(type_ref: TypeRef.T, value: Any, non_null=False):
        match (type_ref, schema.type_map[TypeRef.root_type_name(type_ref)], value):
            # Only allow Null values when non_null=True
            case (_, _, None):
                if not non_null:
                    return InputValue.Null()
                else:
                    raise TypeError(f"Argument {argmeta.name} cannot be None!")

            # If type is non_null, recurse with non_null=True
            case (TypeRef.NonNull(inner=t), _, _):
                return fmt_value(t, value, non_null=True)

            case (TypeRef.Named(name="ID"), _, str()):
                return InputValue.String(value)

            case (TypeRef.Named(name="Int"), _, int()):
                return InputValue.Int(value)
            case (TypeRef.Named(name="BigInt"), _, int()):
                return InputValue.String(str(value))

            case (TypeRef.Named(name="Float"), _, int() | float()):
                return InputValue.Float(float(value))
            case (TypeRef.Named(name="BigDecimal"), _, int() | float()):
                return InputValue.String(str(float(value)))

            case (TypeRef.Named(name="String" | "Bytes"), _, str()):
                return InputValue.String(value)
            case (TypeRef.Named(), TypeMeta.EnumMeta(), str()):
                return InputValue.Enum(value)

            case (TypeRef.Named(name="Boolean"), _, bool()):
                return InputValue.Boolean(value)

            case (TypeRef.List(inner=t), _, list()):
                return InputValue.List([fmt_value(t, val, non_null) for val in value])

            case (TypeRef.Named(), TypeMeta.InputObjectMeta() as input_object, dict()):
                return InputValue.Object(
                    {
                        key: fmt_value(
                            input_object.type_of_input_field(key), val, non_null
                        )
                        for key, val in value.items()
                    }
                )

            case (value, typ, non_null):
                raise TypeError(f"mk_input_value({value}, {typ}, {non_null})")

    return fmt_value(argmeta.type_, value)


def arguments_of_field_args(
    schema: SchemaMeta, field: TypeMeta.FieldMeta, args: dict[str, Any] | None
) -> list[Argument]:
    if args is None:
        args = {}

    def f(arg_meta: TypeMeta.ArgumentMeta) -> Argument | None:
        if arg_meta.name in args:
            return Argument(
                arg_meta.name,
                input_value_of_argument(schema, arg_meta, args[arg_meta.name]),
            )
        else:
            if (arg_meta.default_value) or (not TypeRef.is_non_null(arg_meta.type_)):
                return None
            else:
                raise TypeError(
                    f"arguments_of_field_args: Argument {arg_meta.name} of field"
                    f" {field.name} is required but not provided!"
                )

    # TODO: Add warnings if arguments are not used

    match field:
        case TypeMeta.FieldMeta() as field:
            args = [f(arg_meta) for arg_meta in field.arguments]
            return list(filter(lambda arg: arg is not None, args))
        case _:
            raise TypeError(
                f"arguments_of_field_args: TypeMeta {field.name}"
                " is not of type FieldMeta"
            )



"""
Utility module for Subgrounds
"""

import os
import platform
import warnings
from functools import cache
from itertools import accumulate as _accumulate
from itertools import filterfalse
from operator import itemgetter
from typing import (
    Any,
    Callable,
    Generator,
    Iterator,
    ParamSpec,
    Tuple,
    TypeVar,
    cast,
    overload,
)

from pipe import Pipe, map

PLAYGROUNDS_APP_URL = "https://app.playgrounds.network/"
PLAYGROUNDS_API_URL = "https://api.playgrounds.network/"
PLAYGROUNDS_ENV_VAR = "PLAYGROUNDS_API_KEY"

accumulate = Pipe(_accumulate)


def flatten(t):
    return [item for sublist in t for item in sublist]


def identity(x):
    return x


T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
P = ParamSpec("P")

Container = list[T] | dict[str, T]


def fst(tup: Tuple[T, Any]) -> T:
    return tup[0]


def snd(tup: Tuple[Any, T]) -> T:
    return tup[1]


# ================================================================
# Set utility functions
# ================================================================
def intersection(
    l1: list[T],
    l2: list[T],
    key: Callable[[T], Any] = identity,
    combine: Callable[[T, T], T] = lambda x, _: x,
) -> list[T]:
    l1_keys, l2_keys = list(l1 | map(key)), list(l2 | map(key))
    l2_in_l1 = list(filter(lambda x: key(x) in l1_keys, l2))
    l1_in_l2 = list(filter(lambda x: key(x) in l2_keys, l1))
    l2_in_l1.sort(key=key)
    l1_in_l2.sort(key=key)

    return list(zip(l1_in_l2, l2_in_l1) | map(lambda tup: combine(tup[0], tup[1])))


def rel_complement(
    l1: list[T], l2: list[T], key: Callable[[T], Any] = identity
) -> list[T]:
    l2_keys = list(l2 | map(key))
    return list(filterfalse(lambda x: key(x) in l2_keys, l1))


def sym_diff(l1: list[T], l2: list[T], key: Callable[[T], Any] = identity) -> list[T]:
    return rel_complement(l1, l2, key) + rel_complement(l2, l1, key)


def union(
    l1: list[T],
    l2: list[T],
    key: Callable[[T], Any] = identity,
    combine: Callable[[T, T], T] = lambda x, _: x,
) -> list[T]:
    return (
        rel_complement(l1, l2, key)
        + intersection(l1, l2, key, combine)
        + rel_complement(l2, l1, key)
    )


@overload
def merge(data1: list[T], data2: list[T]) -> list[T]:
    ...


@overload
def merge(data1: dict[str, T], data2: dict[str, T]) -> dict[str, T]:
    ...


def merge(data1: Container, data2: Container) -> Container:
    """Merges ``data1`` and ``data2`` and returns the combined result.

    ``data1`` and ``data2`` must be of the same type. Either both are
    ``dict`` or ``list``.

    >>> a = {'a': 1, 'b': {'c': 2, 'd': 3}}
    >>> b = {'b': {'e': 4}, 'f': 5}
    >>> merge(a, b)
    {'a': 1, 'b': {'c': 2, 'd': 3, 'e': 4}, 'f': 5}

    Args:
      data1 (list[T] | dict[str, T]): First data blob
      data2 (list[T] | dict[str, T]): Second data blob

    Returns:
        list[T] | dict[str, T]: Combined data blob
    """

    match data1, data2:
        case list(l1), list(l2):
            return union(l1, l2, itemgetter("id"), combine=merge)

        case dict(d1), dict(d2):
            data = d1.copy()

            for key in d1:
                if key in d2:
                    data[key] = merge(d1[key], d2[key])

            for key in d2:
                if key not in data:
                    data[key] = d2[key]

            return data

        case (dict(), _) | (_, dict()) | (list(), _) | (_, list()):
            raise TypeError(
                f"merge: incompatible data types!"
                f"type(data1): {type(data1)} != type(data2): {type(data2)}"
            )

        case val1, _:
            return val1


# ================================================================
# Misc
# ================================================================
def filter_none(items: list[T | None]) -> list[T]:
    return list(filter(None, items))


@Pipe
def filter_map(items: Iterator[T], mapping: Callable[[T], T | None]) -> Iterator[T]:
    for item in items:
        new_item = mapping(item)
        if new_item is not None:
            yield new_item


class Sentinel:
    """This class purely used for 'Sentinel' type values such as default arguments
    to functions (where `None` is a significant value) or as a default return (such as
    str.find, where `-1` is returned if a substring isn't found)."""

    def __repr__(self) -> str:
        return "Sentinel"


MISSING = Sentinel()


# ================================================================
# Dictionary related utility functions
# ================================================================
def extract_data(
    keys: list[str], data: dict[str, Any] | list[dict[str, Any]]
) -> list[Any] | Any:
    def f(keys: list[str], data: dict | list | Any):
        match keys:
            case []:
                return data
            case [name, *rest]:
                match data:
                    case dict():
                        return f(rest, data[name])
                    case list():
                        return list(data | map(lambda row: f(rest, row[name])))
                    case None:
                        return None
                    case _:
                        raise Exception(
                            f"extract_data: unexpected state!"
                            f" path = {keys}, data = {data}"
                        )

    match data:
        case dict():
            return f(keys, data)
        case list():
            for doc_data in data:
                try:
                    return f(keys, doc_data)
                except KeyError:
                    continue
            raise Exception(f"extract_data: not found! path = {keys}, data = {data}")
        case _:
            raise Exception("extract_data: data is not dict or list")


def flatten_dict(data: dict[str, Any], keys: list[str] = []) -> dict:
    """Takes a dictionary containing key-value pairs where all values are of type
    other than `list` and flattens it such that all key-value pairs in nested
    dictionaries are now at depth 1.

    Args:
      data Dictionary containing non-list values
      keys: Keys of `data` if `data` is a nested `dict`
        (`len(keys)` == depth of `data`).  Defaults to [].

    Returns:
      dict: Flat dictionary containing all key-value pairs in `data` and its nested
        dictionaries
    """
    flat_dict: dict[str, Any] = {}
    for key, value in data.items():
        match value:
            case dict():
                flat_dict = flat_dict | flatten_dict(value, [*keys, key])
            case value:
                flat_dict["_".join([*keys, key])] = value

    return flat_dict


def contains_list(data: dict[str, Any] | list[Any] | str | int | float | bool) -> bool:
    """Returns `True` if `data` contains a value of type `list` in its nested data
    and `False` otherwise

    Args:
      data (dict | list | str | int | float | bool): Data

    Returns:
      bool: `True` if `data` contains a list, `False` otherwise
    """

    match data:
        case list():
            return True
        case dict():
            return any(data.values() | map(contains_list))
        case set():
            return any(data | map(contains_list))
        case str() | int() | float() | bool():
            return False


def coroutine_generator(
    func: Callable[..., Generator[T | None, U, V]]
) -> Callable[..., Generator[T, U, V]]:
    """This defines a coroutine styled generator.

    All this does is *start* the generator via a `next` call allowing you to use `.send`
     immediately instead of needing to use `gen.send(None)` or `next` first.

    Inspired from: http://www.dabeaz.com/coroutines/Coroutines.pdf (p. 27)

    Essentially, the logic is as follows:
    >>> def start(*args, **kwargs):
    ...     gen = func(*args, **kwargs)
    ...     next(gen)
    ...     return gen
    """

    def start(*args: P.args, **kwargs: P.kwargs) -> Generator[T, U, V]:
        gen = func(*args, **kwargs)
        next(gen)
        return cast(Generator[T, U, V], gen)

    return start


# ================================================================
# User Agent / Headers
# ================================================================


@cache
def default_header(url: str):
    """Contains the default header information for requests made by subgrounds

    Inserts the Playgrounds API Key iff:
      a) targetting the Playgrounds API
      AND
      b) if the `PLAYGROUNDS_API_KEY` environment variable is set
    """

    base = {"Content-Type": "application/json", "User-Agent": user_agent()}

    if url.startswith(PLAYGROUNDS_API_URL) and PLAYGROUNDS_ENV_VAR in os.environ:
        key = os.environ[PLAYGROUNDS_ENV_VAR]
        if key.startswith("pg-"):
            return base | {"Playgrounds-Api-Key": key}

        warnings.warn(
            RuntimeWarning(
                "Invalid Playgrounds Api Key at $PLAYGROUNDS_API_KEY:"
                " Playgrounds api keys should start with 'pg-'.\n\n"
                f"Go to {PLAYGROUNDS_APP_URL} to double check your API Key!"
            )
        )

    return base


@cache
def user_agent():
    """A basic user agent describing the following details:

    - "Subgrounds" (and version)
    - Platform/OS (and architecture)
    - Python Type (and version)

    Examples:
    - Subgrounds/1.1.2 (Darwin; arm64) CPython/3.11.2
    - Subgrounds/1.1.2 (Emscripten; wasm32) CPython/3.10.2

    ⚠️ To override this, construct your :class:`~subgrounds.Subgrounds` with a headers
      parameter with a dictionary containing an empty "User-Agent" key-value pairing.
    """

    from subgrounds import __version__  # avoid import loops

    system = platform.system()
    python = platform.python_implementation()
    python_version = platform.python_version()
    machine = platform.machine()

    return (
        f"Subgrounds/{__version__}"
        f" ({system}; {machine})"
        f" {python}/{python_version}"
    )



from subgrounds.client import AsyncSubgrounds, Subgrounds
from subgrounds.subgraph import FieldPath, Subgraph, SyntheticField

__version__ = "1.9.2"

__all__ = [
    "AsyncSubgrounds",
    "FieldPath",
    "Subgraph",
    "Subgrounds",
    "SyntheticField",
]



""" This module contains the pagination algorithms (both regular and 
iterative) that make use of pagination strategies.
"""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import replace
from typing import Any

from subgrounds.query import Document, DocumentResponse
from subgrounds.schema import SchemaMeta

from .strategies import PaginationStrategy, SkipPagination, StopPagination

PaginateGenerator = Generator[Document, DocumentResponse, None]


class PaginationError(RuntimeError):
    def __init__(self, message: Any, strategy: PaginationStrategy):
        super().__init__(message)
        self.strategy = strategy


def paginate(
    schema: SchemaMeta, doc: Document, pagination_strategy: type[PaginationStrategy]
) -> PaginateGenerator:
    """Evaluates a :class:`PaginationStrategy` against a document based upon a schema.

    Produces a stream of :class:`Document`s until pagination is completed. Documents are
     executed outside the this function and data is transfered via the generator scheme.

    Args:
      schema (SchemaMeta): The GraphQL schema on which the request document is based
      doc (Document): The request document

    Returns:
      dict[str, Any]: The response data as a JSON dictionary
    """

    try:
        strategy = pagination_strategy(schema, doc)
        doc, args = strategy.step()

        while True:
            page = yield replace(doc, variables=doc.variables | args)

            try:
                doc, args = strategy.step(page.data)
            except StopPagination:
                break
            except Exception as exn:
                raise PaginationError(exn.args[0], strategy) from exn

    except SkipPagination:
        _ = yield doc  # consistency



from subgrounds.query import DataRequest, DataResponse, Document, DocumentResponse


class DocumentTransform:
    """Abstract class representing a transformation layer to be applied to
    :class:`Document` objects.
    """

    def transform_document(self, doc: Document) -> Document:
        """Method that will be applied to all :class:`Document` objects that pass
        through the transformation layer.

        Args:
          doc: The initial document

        Returns:
          The transformed document
        """

        return doc

    def transform_response(
        self, req: Document, resp: DocumentResponse
    ) -> DocumentResponse:
        """Method to be applied to all response data ``data`` of requests that pass
        through the transformation layer.

        ``doc`` is the initial :class:`Document` object that yielded the
        resulting JSON data ``data``.

        Args:
          doc: Initial document
          data: Response resulting from the execution of the transformed document.

        Returns:
          The transformed response data
        """

        return resp


class RequestTransform:
    """Abstract class representing a transformation layer to be applied to entire
    :class:`DataRequest` objects.
    """

    def transform_request(self, req: DataRequest) -> DataRequest:
        """Method that will be applied to all :class:`DataRequest` objects that
        pass through the transformation layer.

        Args:
          req: The initial request object

        Returns:
          The transformed request object
        """

        return req

    def transform_response(self, req: DataRequest, resp: DataResponse) -> DataResponse:
        """Method to be applied to all response data ``data`` of requests that pass
        through the transformation layer.

        ``req`` is the initial :class:`DataRequest` object that yielded the
        resulting JSON data ``data``.

        Args:
          req: Initial data request object
          data: Response resulting from the execution of the transformed data request.

        Returns:
          The transformed response data
        """

        return resp



from __future__ import annotations

from typing import overload

from subgrounds.query import InputValue


@overload
def merge_input_value_object_metas(
    data1: InputValue.Object, data2: InputValue.Object
) -> InputValue.Object:
    ...


@overload
def merge_input_value_object_metas(
    data1: dict[str, InputValue.Object], data2: dict[str, InputValue.Object]
) -> dict[str, InputValue.Object]:
    ...


def merge_input_value_object_metas(
    data1: InputValue.Object | dict[str, InputValue.Object],
    data2: InputValue.Object | dict[str, InputValue.Object],
) -> InputValue.Object | dict[str, InputValue.Object]:
    """Merges ``data1`` and ``data2`` and returns the combined result.

    ``data1`` and ``data2`` must be of the same type. Either both are
    ``dict``, ``InputValue.Object``
    """

    match data1, data2:
        case dict(d1), dict(d2):
            data = {}
            for key in d1:
                if key in d2:
                    data[key] = merge_input_value_object_metas(d1[key], d2[key])
                else:
                    data[key] = d1[key]

            for key in d2:
                if key not in data:
                    data[key] = d2[key]

            return data

        case InputValue.Object(value=data1), InputValue.Object(value=data2):
            return InputValue.Object(merge_input_value_object_metas(data1, data2))

        case val1, _:
            return val1



""" `SubgroundsBase`

This module implements the base API that developers can use to build custom clients to
 allow for intricate customization of the subgrounds feature set.

 {{ lab_bdg }} The building of Custom Clients is an experimental feature.
"""

import json
import logging
import warnings
from abc import ABC
from collections.abc import Generator
from dataclasses import dataclass, field
from functools import reduce
from importlib import resources
from pathlib import Path
from typing import Annotated, Any, Type, cast

from pipe import groupby, map, traverse
from typing_extensions import Self

from ..errors import SubgroundsError
from ..pagination import (
    LegacyStrategy,
    PaginationStrategy,
    normalize_strategy,
    paginate,
)
from ..query import DataRequest, DataResponse, Document, DocumentResponse, Query
from ..schema import SchemaMeta
from ..subgraph import FieldPath, Subgraph
from ..transform import (
    DEFAULT_GLOBAL_TRANSFORMS,
    DEFAULT_SUBGRAPH_TRANSFORMS,
    RequestTransform,
    apply_transforms,
)
from ..utils import PLAYGROUNDS_APP_URL

logger = logging.getLogger("subgrounds")
warnings.simplefilter("default")


HTTP2_SUPPORT = True

INTROSPECTION_QUERY = (
    resources.files("subgrounds") / "resources" / "introspection.graphql"
)


@dataclass
class SubgroundsBase(ABC):
    """A base instance for all `Subgrounds` (should not be used directly)"""

    timeout: Annotated[int, "seconds"] = 30
    headers: dict[str, Any] = field(default_factory=dict)
    global_transforms: list[RequestTransform] = field(
        default_factory=lambda: DEFAULT_GLOBAL_TRANSFORMS.copy()
    )
    subgraphs: dict[str, Subgraph] = field(default_factory=dict)
    schema_cache: Path = Path("schemas/")

    def __post_init__(self):
        self.schema_cache = Path(self.schema_cache)

    @classmethod
    def from_pg_key(cls, key: str, **kwargs: Any) -> Self:
        """Create a Subgrounds* instance using a playgrounds key directly.

        This sets the `headers` field internally to be used with all queries made out.

        Args:
          key: The aforementioned Playgrounds API Key
          **kwargs: Anything else to construct the Subgrounds* instance

        Returns:
          An instance Subgrounds* with Playgrounds API support baked in
        """

        if not key.startswith("pg-"):
            raise SubgroundsError(
                "Invalid Playgrounds Key: key should start with 'pg-'.\n\n"
                f"Go to {PLAYGROUNDS_APP_URL} to double check your API Key!"
            )

        headers: dict[str, Any] = {"headers": {"Playgrounds-Api-Key": key}}

        if (headers_arg := kwargs.get("headers")) is not None:
            if "Playgrounds-Api-Key" in headers_arg:
                raise TypeError(
                    f"{cls.__name__}.from_pg_key cannot take `headers`"
                    "as a keyword argument."
                )

            headers |= headers_arg

        return cls(**(kwargs | headers))

    @classmethod
    def _subgraph_slug(cls, url: str) -> str:
        *_, author, name = url.split("/")

        return f"{author}_{name}"

    def make_request(self, fpaths: FieldPath | list[FieldPath]) -> DataRequest:
        """Creates a :class:`DataRequest` object by combining one or more
        :class:`FieldPath` objects.

        Args:
          fpaths: One or more :class:`FieldPath` objects that should be included
           in the request

        Returns:
          Brand new request
        """

        fpaths = list([fpaths] | traverse | map(FieldPath._auto_select) | traverse)

        return DataRequest(
            documents=list(
                fpaths
                | groupby(lambda fpath: fpath._subgraph._url)
                | map(
                    lambda group: Document(
                        url=group[0],
                        query=reduce(
                            Query.add, group[1] | map(FieldPath._selection), Query()
                        ),
                    )
                )
            )
        )

    mk_request = make_request

    def fetch_schema(self, url: str) -> None | dict[str, Any]:
        """Reads schema from filesystem based on subgraph_slug of the url"""

        self.schema_cache.mkdir(parents=True, exist_ok=True)
        schema_path = self.schema_cache / self._subgraph_slug(url)

        if (schema := schema_path.with_suffix(".json")).exists():
            return json.loads(schema.read_text())

    def cache_schema(self, url: str, data: dict[str, Any]):
        """Writes schema into filesystem based on subgraph_slug of the url"""

        self.schema_cache.mkdir(parents=True, exist_ok=True)
        schema_path = self.schema_cache / self._subgraph_slug(url)

        schema = schema_path.with_suffix(".json")
        schema.write_text(json.dumps(data))

    def _load(
        self, url: str, save_schema: bool = False, is_subgraph: bool = True
    ) -> Generator[tuple[str, str], dict[str, Any], Subgraph]:
        """Loads a subgraph / graphql API by fetching the schema.

        If `save_schema` is set `True`, grabs schema from disk and skips query made to
          url.

        ```{note}
        This method utilizes the sans-io generator pattern.
        More on that [here](/subgrounds/advanced_topics/custom_clients/#methodology).
        ```
        """

        if not save_schema or (schema := self.fetch_schema(url)) is None:
            # TODO Yield `Document` once we have graphql -> AST converter
            schema = yield url, INTROSPECTION_QUERY.read_text()

            if save_schema:
                self.cache_schema(url, schema)

        self.subgraphs[url] = Subgraph(
            url,
            SchemaMeta.parse_obj(schema["__schema"]),
            DEFAULT_SUBGRAPH_TRANSFORMS,
            is_subgraph,
        )
        return self.subgraphs[url]

    def _execute(
        self,
        req: DataRequest,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> Generator[Document, DocumentResponse, DataResponse]:
        """Executes a :class:`DataRequest` and returns a :class:`DataResponse`.

        Args:
          req: The :class:`DataRequest` object to be executed.
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
            ``Protocol``. If ``None``, then automatic pagination is disabled.
            Defaults to :class:`LegacyStrategy`.

        Returns:
          A :class:`DataResponse` object representing the response
        """

        strategy = normalize_strategy(pagination_strategy)

        # Setup the main transformation pipeline via `apply_transforms`
        document_transforms = {
            url: subgraph._transforms for url, subgraph in self.subgraphs.items()
        }
        transformer = apply_transforms(self.global_transforms, document_transforms, req)

        # start with the base request and the response we'll be building iteratively
        data_req = cast(DataRequest, next(transformer))
        data_resp = DataResponse(responses=[])

        # for each top-level document (generally 1 per subgraph URL):
        #   define the pagination pipeline
        #   setup the starting doc request and the response we'll be building
        #   iteratively until pagination is complete
        for doc in data_req.documents:
            paginator = paginate(self.subgraphs[doc.url]._schema, doc, strategy)

            doc_req = next(paginator)
            doc_resp = DocumentResponse(url=doc.url, data={})

            while True:
                resp = yield doc_req
                doc_resp = doc_resp.combine(resp)

                try:
                    doc_req = paginator.send(resp)
                except StopIteration:
                    break

            data_resp = data_resp.add_responses(doc_resp)

        next(transformer)  # toss empty None
        return cast(DataResponse, transformer.send(data_resp))

    def _execute_iter(
        self,
        req: DataRequest,
        pagination_strategy: Type[PaginationStrategy] | None = LegacyStrategy,
    ) -> Generator[Document | DocumentResponse, DocumentResponse, None]:
        """Same as `execute`, except that an iterator is returned which will iterate
        the data pages.

        Args:
          req: The :class:`DataRequest` object to be executed
          pagination_strategy: A Class implementing the :class:`PaginationStrategy`
           ``Protocol``. If ``None``, then automatic pagination is disabled.
           Defaults to :class:`LegacyStrategy`.

        Returns:
          An iterator over the :class:`DocumentResponse` pages.

        ⚠️ DOES NOT apply global transforms across multiple documents or their pages.
         Since we yield each page as we get it, it's not possible to accurately perform
         the transforms since we don't collect the pages. This means transforms
         expecting multiple documents or pages of documents will be inaccurate.
        """

        if len(req.documents) > 1:
            raise NotImplementedError(
                "`execute_iter` cannot currently handle multiple documents\n"
                "Try splitting your query into multiple queries."
            )

        document_transforms = {
            url: subgraph._transforms for url, subgraph in self.subgraphs.items()
        }
        transformer = apply_transforms(self.global_transforms, document_transforms, req)
        strategy = normalize_strategy(pagination_strategy)

        data_req = cast(DataRequest, next(transformer))
        for doc in data_req.documents:
            paginator = paginate(self.subgraphs[doc.url]._schema, doc, strategy)
            paginated_doc = next(paginator)
            while True:
                resp = yield paginated_doc

                next(transformer)  # toss empty None
                data_resp = cast(
                    DataResponse, transformer.send(DataResponse(responses=[resp]))
                )
                yield data_resp.responses[0]  # will only be one

                try:
                    paginated_doc = paginator.send(resp)
                except StopIteration:
                    break



""" Subgrounds request/response transformation layers module

This module defines interfaces (abstract classes) for transformation layers.
Transformation layers, or transforms, can be applied to entire
requests (see :class:`RequestTransform`) or on a per-document basis (see
:class:`DocumentTransform`). Classes that implement either type of transforms
can be used to perform modifications to queries and their response data.

For example, the :class:`TypeTransform` class is used to tranform the response
data of ``BigInt`` and ``BigDecimal`` fields (which are represented as
strings in the response JSON data) to python ``int`` and ``float``
respectively (see the actual transforms in ``DEFAULT_SUBGRAPH_TRANSFORMS``).

Transforms are also used to apply :class:`SyntheticField` to queries and the
response data (see :class:`LocalSyntheticField` transform class). Each
:class:`SyntheticField` defined on a subgraph creates a new transformation layer
by instantiating a new :class:`LocalSyntheticField` object.
"""

import logging
from functools import partial
from typing import Any

from pipe import map, traverse

from subgrounds.query import Selection
from subgrounds.schema import TypeMeta

logger = logging.getLogger("subgrounds")


def select_data(select: Selection, data: dict) -> list[Any]:
    match (select, data):
        case (
            Selection(TypeMeta.FieldMeta(name=name), None, _, [] | None)
            | Selection(TypeMeta.FieldMeta(), name, _, [] | None),
            dict() as data,
        ) if name in data:
            return [data[name]]

        case (
            Selection(TypeMeta.FieldMeta(name=name), None, _, inner_select)
            | Selection(TypeMeta.FieldMeta(), name, _, inner_select),
            dict() as data,
        ) if name in data:
            return list(
                inner_select | map(partial(select_data, data=data[name])) | traverse
            )

        case (select, data):
            raise Exception(f"select_data: invalid selection {select} for data {data}")

    assert False  # Suppress mypy missing return statement warning



"""Subgrounds Dash Components

Extending dash components to be able to understand subgrounds logic. This includes other
  extended components of other libraries such as `plotly`.
"""

from .abcs import Refreshable
from .components import AutoUpdate, DataTable, Graph

__all__ = [
    "Refreshable",
    "Graph",
    "DataTable",
    "AutoUpdate",
]



"""This module contains all code related to automatic pagination.

The `pagination` module contains the pagination algorithms (both regular and iterative)
  that make use of `PaginationStrategies`.

The `preprocess` and `strategies` modules implement the currently supported
  `PaginationStrategies`: `LegacyStrategy` and `ShallowStrategy`.

The `utils` module contains some generic functions that are useful in the context of
  pagination.
"""

from .pagination import PaginationError, PaginationStrategy, paginate
from .preprocess import PaginationNode, generate_pagination_nodes, prune_doc
from .strategies import (
    LegacyStrategy,
    ShallowStrategy,
    SkipStrategy,
    normalize_strategy,
)

__all__ = [
    "generate_pagination_nodes",
    "LegacyStrategy",
    "normalize_strategy",
    "paginate",
    "paginate",
    "PaginationError",
    "PaginationNode",
    "PaginationStrategy",
    "prune_doc",
    "ShallowStrategy",
    "SkipStrategy",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from dash.dependencies import Output


class Refreshable(ABC):
    @property
    @abstractmethod
    def dash_dependencies(self) -> list[Output]:
        raise NotImplementedError

    @property
    @abstractmethod
    def dash_dependencies_outputs(self) -> list[Any]:
        raise NotImplementedError
